# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import tempfile, shutil, os
from collections import namedtuple
import unittest
from railroadtracks import rnaseq, easy
from railroadtracks.easy import tasksetgraph
import railroadtracks.model.simulate
PHAGEFASTA = railroadtracks.model.simulate.PHAGEFASTA

class TaskSetGraphTestCase(unittest.TestCase):
    def setUp(self):
        wd = tempfile.mkdtemp()
        self.wd2 = tempfile.mkdtemp()
        self.project = easy.Project(rnaseq, wd)
        bowtie2index = rnaseq.Bowtie2Build()
        reference_fn = PHAGEFASTA
        Assets = bowtie2index.Assets
        task = self.project.add_task(bowtie2index, 
                                     Assets(Assets.Source(rnaseq.FASTAFile(reference_fn))))
        self.task = task


    def tearDown(self):
        shutil.rmtree(self.project.wd)
        shutil.rmtree(self.wd2)

    def _create_fastq(self, n, d):
        reference_fn = PHAGEFASTA
        with open(reference_fn) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))

        ReadPair = namedtuple('ReadPair', 'read1 read2')
        pairnames = tuple(ReadPair(os.path.join(d, 'A_%i_1.fq' % p_i), 
                                   os.path.join(d, 'A_%i_2.fq' % p_i)) for p_i in range(1, n+1))
        for read1_fn, read2_fn in pairnames:
            with open(read1_fn, 'wb') as read1_fh, \
                 open(read2_fn, 'wb') as read2_fh:
                railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh, reference)
        return pairnames


    def testAddTaskSet(self):
        ts = easy.TaskSet()
        ts.add(self.task)
        #
        tsg = tasksetgraph.TaskSetGraph()
        tsindex_id = tsg.add_taskset(ts)
        dict_retrieve = tsg.digraph.node[tasksetgraph.get_tasksetid((self.task,))]
        ts_retrieve = dict_retrieve['taskset']
        self.assertEqual(len(ts), len(ts_retrieve))
        self.assertEqual(tuple(x.task_id for x in ts), tuple(x.task_id for x in ts_retrieve))

        # add connected tasks
        taskset_align = easy.TaskSet()
        aligner = rnaseq.Bowtie2()
        for fq_name in ('foo.fq', 'bar.fq'):
            Assets = aligner.Assets
            assets = Assets(Assets.Source(self.task.call.assets.target.indexfilepattern,
                                          rnaseq.FASTQPossiblyGzipCompressed(fq_name),
                                          None), None)
            task_align = self.project.add_task(aligner, assets)
            taskset_align.add(task_align)
        tsalign_id = tsg.add_taskset(taskset_align)
        lst = tsg.execution_list()
        self.assertEqual(tsindex_id, tasksetgraph.get_tasksetid(lst[0].taskset))
        self.assertEqual(tsalign_id, tasksetgraph.get_tasksetid(lst[1].taskset))

    def testAddTaskSetDuplicateTask(self):
        ts = easy.TaskSet()
        ts.add(self.task)
        tsg = tasksetgraph.TaskSetGraph()
        tsg.add_taskset(ts)

        ts2 = easy.TaskSet()
        ts2.add(self.task)
        self.assertRaises(ValueError, tsg.add_taskset, ts2)

    def testAddTask(self):
        tsg = tasksetgraph.TaskSetGraph()
        tsg.add(self.task)
        dict_retrieve = tsg.digraph.node[tasksetgraph.get_tasksetid((self.task,))]
        ts_retrieve = dict_retrieve['taskset']
        self.assertEqual(1, len(ts_retrieve))
        self.assertEqual((self.task.task_id,), tuple(x.task_id for x in ts_retrieve))


    def testAddTaskDifferentProject(self):
        project2 = easy.Project(rnaseq, self.wd2)
        bowtie2index = rnaseq.Bowtie2Build()
        reference_fn = PHAGEFASTA
        Assets = bowtie2index.Assets
        task2 = project2.add_task(bowtie2index, 
                                 Assets(Assets.Source(rnaseq.FASTAFile(reference_fn))))
        #
        tsg = tasksetgraph.TaskSetGraph()
        tsg.add(self.task)
        self.assertRaises(ValueError, tsg.add, task2)

    def testExecutionList(self):
        tsg = tasksetgraph.TaskSetGraph()
        # initial task/taskset is building a bowtie2 index
        tsindex_id = tsg.add(self.task)
        lst = tsg.execution_list()
        self.assertEqual(1, len(lst))

        # add a second task set that performs the alignment of several pairs of
        # FASTQ files
        n_fastq = 3
        pairnames = self._create_fastq(n_fastq, self.wd2)

        index_task = self.task
        bowtie2 = rnaseq.Bowtie2()
        Assets = bowtie2.Assets
        # Class to model FASTQ files that are optionally compressed
        FASTQ = rnaseq.FASTQPossiblyGzipCompressed
        ts = easy.TaskSet()
        # note that we are included the pair of FASTQ already aligned above
        for read1_fn, read2_fn in pairnames:
            task = self.project.add_task(bowtie2,
                                         Assets(Assets.Source(index_task.call.assets.target.indexfilepattern,
                                                              FASTQ(read1_fn),
                                                              FASTQ(read2_fn))))
            ts.add(task)
        tsalign_id = tsg.add(ts)
        lst = tsg.execution_list()

    def testExecution(self):
        tsg = tasksetgraph.TaskSetGraph()
        tsg.add(self.task)
        n_fastq = 3
        pairnames = self._create_fastq(n_fastq, self.wd2)

        index_task = self.task
        bowtie2 = rnaseq.Bowtie2()
        Assets = bowtie2.Assets
        FASTQ = rnaseq.FASTQPossiblyGzipCompressed
        ts = easy.TaskSet()
        for read1_fn, read2_fn in pairnames:
            task = self.project.add_task(bowtie2,
                                         Assets(Assets.Source(index_task.call.assets.target.indexfilepattern,
                                                              FASTQ(read1_fn),
                                                              FASTQ(read2_fn))))
            ts.add(task)
        tsg.add(ts)
        # no task mapper, check that an exception is raised
        self.assertRaises(ValueError, tsg.execute)

        # set the default task mapper to a multiprocessing based one
        n_processes = 1
        mpe = easy.MultiprocessingExecution(n_processes)
        tsg.defaultmapper = mpe
        tsg.execute()

        # use a run-only-once filter (that is only try to execute
        # tasks with the status "TO DO"
        flt = lambda taskset: taskset.filter_on_status(easy._TASK_TODO)
        tsg.defaulttasksetfilter = flt
        lst = tsg.execution_list()
        for elt in lst:
            ts_f = tsg._filtered_taskset(elt)
            # check that nothing is left to be exectuted
            # (since everything has been run earlier)
            self.assertEqual(0, len(ts_f))
    



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile, shutil, os, subprocess, csv, random, re, itertools, gzip
import unittest
from railroadtracks import environment, unifex, core, rnaseq

from railroadtracks.model.simulate import (PHAGEGFF,
                                           PHAGEGTF,
                                           PHAGEFASTA)
import railroadtracks.model.aligners
import railroadtracks.model.simulate
import railroadtracks.model.diffexp
import railroadtracks.model.quantify

try:
    import ngs_plumbing
    has_ngsp = True
except ImportError:
    has_ngsp = False


def _build_StepIndex(cls, executable, 
                     tempdir, parameters=()):
    """ 
    :param cls: class used to model the step
    :param executable: name (or path) of executable 
    :param tempdir: temporary directory 

    :return (assets, cmd, returncode):"""
    runner = cls(executable)
    AssetsIndexer = railroadtracks.model.aligners.AssetsIndexer
    referencegenome = rnaseq.FASTAFile(PHAGEFASTA)
    source = AssetsIndexer.Source(referencegenome)
    indexthing = rnaseq.FilePattern(tempdir)
    target = AssetsIndexer.Target(indexthing)
    assets = railroadtracks.model.aligners.AssetsIndexer(source,
                                                         target)
    cmd, returncode = runner.run(assets, parameters = parameters)
    return (assets, cmd, returncode)


def _build_StepAlign(assets_index,
                     cls_align, executable_align,
                     read1_fh):
    runner = cls_align(executable_align)
    AssetsAligner = railroadtracks.model.aligners.AssetsAligner
    # single reads
    read1 = rnaseq.FASTQPossiblyGzipCompressed(read1_fh.name)
    source = AssetsAligner.Source(assets_index.target.indexfilepattern,
                                  read1, None)
    field_i = AssetsAligner.Target._fields.index('alignment')
    cls_alignedreads = getattr(AssetsAligner.Target, core.AssetMetaReserved.SOURCES.value)[field_i].cls
    if cls_alignedreads is rnaseq.SAMFile:
        suffix='.sam'
    elif cls_alignedreads is rnaseq.BAMFile:
        suffix='.bam'
    else:
        raise ValueError('Invalid class for alignments')
    alignment_fh = tempfile.NamedTemporaryFile(suffix=suffix)
    alignment_fh.close()
    alignthing = cls_alignedreads(alignment_fh.name)
    target = AssetsAligner.Target(alignthing)
    assets = AssetsAligner(source,
                           target)
    cmd, returncode = runner.run(assets)
    return (runner.version, 
            assets, cmd, returncode, alignment_fh)

def _build_UpToAlign(cls_index, executable_index,
                     cls_align, executable_align,
                     read1_fh, tempdir,
                     parameters_index = (),
                     align_max_attempts = 1):
    assets_index, cmd, returncode = _build_StepIndex(cls_index, executable_index,
                                                     tempdir, 
                                                     parameters = parameters_index)
    for n_attempt in range(1, align_max_attempts+1):
        try:
            version, assets, cmd, returncode, fh = _build_StepAlign(assets_index,
                                                                    cls_align, executable_align,
                                                                    read1_fh)
            break
        except AssertionError as ae:
            if n_attempt == align_max_attempts:
                raise(ase)

    return (assets, cmd, returncode, fh)


def _build_StepQuantify(assets_align,
                        cls_count, executable_count,
                        parameters_count):
    #FIXME: should be sorted by ID !
    runner = cls_count(executable_count)
    AssetsQuantifier = quantify.AssetsQuantifier
    alignmentfiles = tuple(assets_align.target.alignment)
    #FIXME: len == 1 (tested elsewhere)
    f_align = rnaseq.BAMFile(os.path.join(os.path.dirname(assets_align.target.alignment.name),
                                          alignmentfiles[0]))
    referenceannotation = rnaseq.GFFFile(PHAGEGFF)
    source = AssetsQuantifier.Source(f_align,
                                     referenceannotation)
    fh = tempfile.NamedTemporaryFile(suffix='.csv')
    counts = rnaseq.CSVFile(fh.name)
    target = AssetsQuantifier.Target(counts)
    assets = AssetsQuantifier(source, target)
    cmd, returncode = runner.run(assets, parameters = parameters_count)
    return (assets, cmd, returncode, fh) # fh returned to be protected from premature destruction
    


class ModelSimulateTestCase(unittest.TestCase):

    def setUp(self):
        self._dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self._dir)

    def _test_FluxsimulatorExpression(self, fse):
        A = fse.Assets
        annotation_fn = PHAGEGTF
        genome_dir = os.path.dirname(annotation_fn)

        params_fn = os.path.join(self._dir, 'original.par')
        with open(params_fn, 'w') as fh_out:
            fh_out.write("""
FRAG_SUBSTRATE  RNA
FRAG_METHOD     UR
""")

        paramsandpro_prefix = os.path.join(self._dir, 'modified')
        assets = A(A.Source(rnaseq.GFFFile(annotation_fn),
                            rnaseq.FilePattern(genome_dir),
                            rnaseq.FluxSimulatorParameters(params_fn)), 
                   A.Target(rnaseq.FluxSimulatorParametersAndPro(paramsandpro_prefix)))
        return assets

    @unittest.skipIf(not environment.Executable.ispresent('flux-simulator'),
                     'flux-simulator is not in the PATH')
    def test_FluxsimulatorExpression(self):
        fse = rnaseq.FluxsimulatorExpression()
        assets = self._test_FluxsimulatorExpression(fse)
        cmd, returncode = fse.run(assets, parameters=('--NB-MOLECULES', '1000'))
        self.assertEqual(0, returncode)
        # check that the target file is not empty
        self.assertGreater(os.stat(assets.target.parameters_and_pro.name + '.pro').st_size, 0)

    @unittest.skipIf(not environment.Executable.ispresent('flux-simulator'),
                     'flux-simulator is not in the PATH')
    def test_FluxsimulatorDerivedExpression(self):
        fse = rnaseq.FluxsimulatorExpression()
        assets = self._test_FluxsimulatorExpression(fse)
        cmd, returncode = fse.run(assets, parameters=('--NB-MOLECULES', '1000'))
        # now devire a new expression profile
        fse = rnaseq.FluxsimulatorDerivedExpression()
        A = fse.Assets
        paramsandpro_prefix_source = os.path.join(self._dir, 'modified')
        paramsandpro_prefix_target = os.path.join(self._dir, 'modified_again')
        assets = A(A.Source(rnaseq.FluxSimulatorParametersAndPro(paramsandpro_prefix_source)),
                   A.Target(rnaseq.FluxSimulatorParametersAndPro(paramsandpro_prefix_target)))
        cmd, returncode = fse.run(assets)
        
        self.assertEqual(0, returncode)
        # check that the target file is not empty
        self.assertGreater(os.stat(assets.target.parameters_and_pro.name + '.pro').st_size, 0)

    @unittest.skipIf(not environment.Executable.ispresent('flux-simulator'),
                     'flux-simulator is not in the PATH')
    def test_FluxsimulatorSequencing(self):
        # expression
        fse = rnaseq.FluxsimulatorExpression()
        fse_assets = self._test_FluxsimulatorExpression(fse)
        cmd, returncode = fse.run(fse_assets, parameters=('--NB-MOLECULES', '1000'))
        read1_fn = os.path.join(self._dir, 'read1.fq.gz')
        read2_fn = os.path.join(self._dir, 'read2.fq.gz')

        fss = rnaseq.FluxsimulatorSequencing()
        A = fss.Assets
        paramsandpro_prefix = os.path.join(self._dir, 'expr_modified')
        lib_fn = os.path.join(self._dir, 'mapping.lib')
        #annotation_fn = fse_assets.source.annotation_gtf
        assets = A(A.Source(fse_assets.target.parameters_and_pro),
                   A.Target(rnaseq.FluxSimulatorParametersAndPro(paramsandpro_prefix),
                            rnaseq.FASTQPossiblyGzipCompressed(read1_fn),
                            rnaseq.FASTQPossiblyGzipCompressed(read2_fn),
                            core.File(lib_fn)))
        cmd, returncode = fss.run(assets, parameters=('--read-number', '1000'))
        self.assertEqual(0, returncode)
        # check that the target files are not empty
        self.assertGreater(os.stat(assets.target.parameters_and_pro.name + '.pro').st_size, 0)
        self.assertGreater(os.stat(read1_fn).st_size, 0)
        self.assertGreater(os.stat(read2_fn).st_size, 0)
        self.assertGreater(os.stat(lib_fn).st_size, 0)



class ModelUtilsTestCase(unittest.TestCase):

    def setUp(self):
        cls_index = rnaseq.Bowtie2Build
        executable_index = 'bowtie2-build'
        cls_align = rnaseq.Bowtie2
        executable_align = 'bowtie2'
        read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq')
        read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq')
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, 
                                                                                     read2_fh, 
                                                                                     reference,
                                                                                     n = 500)
        self.tempdir = tempfile.mkdtemp()
        assets_align, cmd, returncode, fh = _build_UpToAlign(cls_index, executable_index,
                                                             cls_align, executable_align, self._read1_fh,
                                                             os.path.join(self.tempdir, 'reference'))
        self._assets_align = assets_align
        self._fh = fh # protect temp file from deletion

    def tearDown(self):
        self._read1_fh.close()
        self._read2_fh.close()
        shutil.rmtree(self.tempdir)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_ensuresam(self):
        assets_align = self._assets_align
        fn = assets_align.target.alignment.name
        self.assertTrue(fn.endswith('.bam'))
        fh_sam = railroadtracks.model.files.ensure_sam(fn)
        self.assertTrue(fh_sam.name.endswith('.sam'))
        fh_sam_again = railroadtracks.model.files.ensure_sam(fh_sam.name)
        self.assertEqual(fh_sam.name, fh_sam_again.name)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_ensurebam(self):
        assets_align = self._assets_align
        fn = assets_align.target.alignment.name
        self.assertTrue(fn.endswith('.bam'))
        fh_sam = railroadtracks.model.files.ensure_sam(fn)
        fh_bam = railroadtracks.model.files.ensure_bam(fh_sam.name)
        fh_bam_again = railroadtracks.model.files.ensure_bam(fh_bam.name)
        self.assertEqual(fh_bam.name, fh_bam_again.name)
        self.assertTrue(fh_bam.name.endswith('.bam'))
        # this is round trip for the SAM file
        with open(fn, mode='rb') as fh_bam_orig:
            self.assertEqual(fh_bam_orig.read(), fh_bam.read())

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_SamtoolsBamToSam(self):
        executable = 'samtools'
        bamtosam = railroadtracks.model.files.SamtoolsBamToSam(executable)
        version = bamtosam.version # not a test - just here to check that the accessor is working
        Assets = bamtosam.Assets
        assets_align = self._assets_align
        source = Assets.Source(assets_align.target.alignment)
        sam_fh = tempfile.NamedTemporaryFile(suffix='.sam')
        target = Assets.Target(rnaseq.SAMFile(sam_fh.name))
        assets = Assets(source, target)
        cmd, returncode = bamtosam.run(assets)
        self.assertEqual(0, returncode)
        # check the infamous '.bam.bam' issue
        self.assertFalse(os.path.exists(sam_fh.name + '.sam')) # if this fails, a newer version of samtools is required
        # check that the target file is not empty
        self.assertGreater(os.stat(sam_fh.name).st_size, 0)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_SamtoolsSamToBam(self):
        executable = 'samtools'
        bamtosam = railroadtracks.model.files.SamtoolsBamToSam(executable)
        Assets = bamtosam.Assets
        assets_align = self._assets_align
        sam_fh = tempfile.NamedTemporaryFile(suffix='.sam')
        assets_bamtosam = Assets(Assets.Source(assets_align.target.alignment), 
                                 Assets.Target(rnaseq.SAMFile(sam_fh.name)))
        cmd, returncode = bamtosam.run(assets_bamtosam)

        samtobam = railroadtracks.model.files.SamtoolsSamToBam(executable)
        version = samtobam.version # not a test - just here to check that the accessor is working
        Assets = samtobam.Assets
        bam_fh = tempfile.NamedTemporaryFile(suffix='.bam')
        assets_samtobam = Assets(Assets.Source(assets_bamtosam.target.samfile), 
                                 Assets.Target(rnaseq.BAMFile(bam_fh.name)))
        cmd, returncode = samtobam.run(assets_samtobam)
        self.assertEqual(0, returncode)
        # check that the target file is not empty
        self.assertGreater(os.stat(bam_fh.name).st_size, 0)




class ModelTestCase(unittest.TestCase):

    def test__make_stepdict(self):
        class MyStep(core.StepAbstract):
            _name = 'My Step'
            _default_execpath = None
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            Assets = None
            version = '0.1'
            def run(self):
                pass
        class MyModule(object):
            pass
        m = MyModule()
        m.Mystep = MyStep
        sl = unifex._make_stepdict(m)
        self.assertEqual(1, len(sl))
        class MyOtherStep(core.StepAbstract):
            _name = 'My Other Step'
            _default_execpath = None
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            Assets = None
            version = '0.1'
            def run(self):
                pass

        m = MyModule()
        m.Mystep = MyStep
        m.MyOtherStep = MyOtherStep
        sl = unifex._make_stepdict(m)
        self.assertEqual(2, len(sl))
        class MyYetAnotherStep(core.StepAbstract):
            _name = 'My Other Step'
            _default_execpath = None
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            Assets = None
            version = '0.1'
            def run(self):
                pass
        m = MyModule()
        m.Mystep = MyStep
        m.MyOtherStep = MyOtherStep
        m.MyYetAnotherStep = MyYetAnotherStep
        self.assertRaises(ValueError, unifex._make_stepdict, m)

    # note: testing core.StepAbstract
    def test_newStepAbstract(self):
        # dummy step that uses the UNIX command 'cat' to concatenate files
        class CatStep(core.StepAbstract):
            _name = 'cat'
            _default_execpath = 'cat'
            Assets = None
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            def getversion(self):
                # dummy version that is the time stamp for the executable
                path = subprocess.check_output(('which', 'cat'))
                ctime = os.path.ctime(path)
                return str(ctime)
            version = property(getversion)
            def run(self, sources, targets, parameters):
                output_filename = targets['output'][0]
                with open(output_filename, 'w') as output:
                    args = ['cat', ]
                    args.extend(sources['input'])
                    res = subprocess.check_call(args, stdout=output)
                return res

        cs = CatStep()
        inputs = (tempfile.NamedTemporaryFile(mode='w+'), 
                  tempfile.NamedTemporaryFile(mode='w+'))
        inputs[0].write('abc')
        inputs[0].flush()
        inputs[1].write('def')
        inputs[1].flush()
        outputs = (tempfile.NamedTemporaryFile(),)
        sources = {'input': tuple(x.name for x in inputs, )}
        targets = {'output': tuple(x.name for x in outputs)}
        res = cs.run(sources, targets, '')
        out_res = outputs[0].readlines()
        self.assertEqual([b'abcdef'], out_res)


class ModelIndexTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_StepBowtie2Build(self):
        execname = 'bowtie2-build'
        runner = rnaseq.Bowtie2Build(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, bytes))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(rnaseq.Bowtie2Build, execname,
                                                   os.path.join(self.tempdir, 'reference'))
        self.assertEqual(0, returncode)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie-build'),
                     'bowtie-build is not in the PATH')
    def test_StepBowtieBuild(self):
        execname = 'bowtie-build'
        runner = rnaseq.BowtieBuild(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, str))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(rnaseq.BowtieBuild, execname,
                                                   os.path.join(self.tempdir, 'reference'))
        self.assertEqual(0, returncode)
        # check that files actually present
        files = tuple(assets.target.indexfilepattern.iterlistfiles())
        self.assertGreater(len(files), 0)

    @unittest.skipIf(not environment.Executable.ispresent('bwa'),
                     'bwa is not in the PATH')
    def test_StepBWAIndex(self):
        execname = 'bwa'
        runner = rnaseq.BWAIndex(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, bytes))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(rnaseq.BWAIndex, execname,
                                                   os.path.join(self.tempdir, 'reference'))
        self.assertEqual(0, returncode)
        # check that files actually present
        files = tuple(assets.target.indexfilepattern.iterlistfiles())
        self.assertGreater(len(files), 0)

    @unittest.skipIf(not environment.Executable.ispresent('STAR'),
                     'STAR is not in the PATH')
    def test_StepStarIndex(self):
        execname = 'STAR'
        runner = rnaseq.StarIndex(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, str))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(rnaseq.StarIndex, execname, 
                                                   os.path.join(self.tempdir, 'reference'),
                                                   parameters=('--genomeChrBinNbits', '12',
                                                               '--genomeSAindexNbases', '8'))
        self.assertEqual(0, returncode)

    @unittest.skipIf(not environment.Executable.ispresent('gmap_build'),
                     'gmap_build is not in the PATH')
    def test_StepGsnapIndex(self):
        execname = 'gmap_build'
        runner = rnaseq.GsnapIndex(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, str))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(rnaseq.GsnapIndex, execname,
                                                   os.path.join(self.tempdir, 'reference'),
                                                   parameters = ('-k', '8'))
        self.assertEqual(0, returncode)

    @unittest.skipIf(not environment.Executable.ispresent('sailfish'),
                     'sailfish is not in the PATH')
    def test_StepSailfishIndex(self):
        execname = 'sailfish'
        runner = rnaseq.SailfishIndex(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, str))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(rnaseq.SailfishIndex, execname, 
                                                   os.path.join(self.tempdir, 'reference'),
                                                   parameters = rnaseq.SailfishIndex.PARAMETERS_DEFAULT)
        self.assertEqual(0, returncode)

    @unittest.skipIf(not environment.Executable.ispresent('salmon'),
                     'salmon is not in the PATH')
    def test_StepSalmonIndex(self):
        execname = 'salmon'
        SalmonIndex = railroadtracks.model.aligners.SalmonIndex
        runner = SalmonIndex(execname)
        #FIXME: test version
        version = runner.version
        self.assertTrue(isinstance(version, str))
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.INDEX, )))
        assets, cmd, returncode = _build_StepIndex(SalmonIndex, execname, 
                                                   os.path.join(self.tempdir, 'reference'),
                                                   parameters = ('--threads', '1'))
        self.assertEqual(0, returncode)


class ModelAlignTestCase(unittest.TestCase):

    def setUp(self):
        NFRAGMENTS_NOMATCH = 10
        read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq')
        read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq')
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, 
                                                                                     read2_fh, 
                                                                                     reference,
                                                                                     n=500)
        # reads from a random reference (should not match)
        randint = railroadtracks.model.simulate.random.randint
        rand_reference = railroadtracks.model.simulate.Entry('> Random DNA',
                                                             bytearray(''.join('ATCG'[randint(0, 3)] for x in range(500)), 'ascii'))
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh,
                                                                                     rand_reference,
                                                                                     n = NFRAGMENTS_NOMATCH)
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        self._read1_fh.close()
        self._read2_fh.close()
        shutil.rmtree(self.tempdir)

    def _test_StepAlign_noreads(self, 
                                cls_index, executable_index,
                                cls_align, executable_align,
                                parameters_index = ()):
        assets_index, cmd, index_res = _build_StepIndex(cls_index, executable_index,
                                                        os.path.join(self.tempdir, 'reference'),
                                                        parameters = parameters_index)
        runner = cls_align(executable_align)
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.ALIGN, )))
        AssetsAligner = railroadtracks.model.aligners.AssetsAligner
        # no reads
        self.assertRaises(AssertionError,
                          AssetsAligner.Source,
                          assets_index.target.indexfilepattern)

    def _test_StepAlign_singlereads(self, 
                                    cls_index, executable_index,
                                    cls_align, executable_align,
                                    parameters_index = (),
                                    align_max_attempts = 1):
        runner = cls_align(executable_align)
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.ALIGN, )))
        # single reads
        assets, cmd, returncode, fh = _build_UpToAlign(cls_index, executable_index,
                                                       cls_align, executable_align, self._read1_fh,
                                                       os.path.join(self.tempdir, 'reference'),
                                                       parameters_index = parameters_index,
                                                       align_max_attempts = align_max_attempts)
        self.assertTrue(isinstance(runner.version, bytes))
        self.assertEqual(0, returncode)
        # FIXME: check that the alignment file contains what it should
        self.assertTrue(os.path.exists(fh.name))
        # check that the target file is not empty
        self.assertGreater(os.stat(fh.name).st_size, 0)

    def _test_StepAlign_pairedreads(self, 
                                    cls_index, executable_index,
                                    cls_align, executable_align,
                                    parameters_index = (),
                                    align_max_attempts = 1):
        assets_index, cmd, returncode = _build_StepIndex(cls_index, executable_index,
                                                         os.path.join(self.tempdir, 'reference'),
                                                         parameters = parameters_index)
        runner = cls_align(executable_align)
        self.assertEqual(set(runner.activities), set((rnaseq.ACTIVITY.ALIGN, )))
        AssetsAligner = railroadtracks.model.aligners.AssetsAligner
        # paired-end reads
        read1 = rnaseq.FASTQPossiblyGzipCompressed(self._read1_fh.name)
        read2 = rnaseq.FASTQPossiblyGzipCompressed(self._read2_fh.name)
        source = AssetsAligner.Source(assets_index.target.indexfilepattern,
                                      read1, read2)
        field_i = AssetsAligner.Target._fields.index('alignment')
        cls_alignedreads = getattr(AssetsAligner.Target, core.AssetMetaReserved.SOURCES.value)[field_i].cls
        if cls_alignedreads is rnaseq.SAMFile:
            suffix='.sam'
        elif cls_alignedreads is rnaseq.BAMFile:
            suffix='.bam'
        else:
            raise ValueError('Invalid class for alignments')
        fh = tempfile.NamedTemporaryFile(suffix=suffix)
        fh.close()
        try:
            alignthing = cls_alignedreads(fh.name)
            target = AssetsAligner.Target(alignthing)
            assets = AssetsAligner(source,
                                   target)
            for n_attempt in range(1, align_max_attempts+1):
                cmd, returncode = runner.run(assets)
                if returncode == 0:
                    break
            # FIXME: check that the alignment file contains what it should
            self.assertEqual(0, returncode)
            self.assertTrue(os.path.exists(fh.name))
            # check that the target file is not empty
            self.assertGreater(os.stat(fh.name).st_size, 0)
        finally:
            if os.path.exists(fh.name):
                os.unlink(fh.name)


    @unittest.skipIf(not environment.Executable.ispresent('bwa'),
                     'bwa is not in the PATH')
    def test_StepBWA_noread(self):
        self._test_StepAlign_noreads(rnaseq.BWAIndex, 'bwa',
                                     rnaseq.BWA, 'bwa')
    @unittest.skipIf(not environment.Executable.ispresent('bwa'),
                     'bwa is not in the PATH')
    def test_StepBWA_singlereads(self):
        self._test_StepAlign_singlereads(rnaseq.BWAIndex, 'bwa',
                                         rnaseq.BWA, 'bwa')
    @unittest.skipIf(not environment.Executable.ispresent('bwa'),
                     'bwa is not in the PATH')
    def test_StepBWA_pairedreads(self):
        self._test_StepAlign_pairedreads(rnaseq.BWAIndex, 'bwa',
                                         rnaseq.BWA, 'bwa')

       
    @unittest.skipIf(not environment.Executable.ispresent('bowtie'),
                     'bowtie is not in the PATH')
    def test_StepBowtie_noread(self):
        self._test_StepAlign_noreads(rnaseq.Bowtie2Build, 'bowtie-build',
                                     rnaseq.Bowtie2, 'bowtie')
    @unittest.skipIf(not environment.Executable.ispresent('bowtie'),
                     'bowtie is not in the PATH')
    def test_StepBowtie_singlereads(self):
        self._test_StepAlign_singlereads(rnaseq.BowtieBuild, 'bowtie-build',
                                         rnaseq.Bowtie, 'bowtie')
    @unittest.skipIf(not environment.Executable.ispresent('bowtie'),
                     'bowtie is not in the PATH')
    def test_StepBowtie_pairedreads(self):
        self._test_StepAlign_pairedreads(rnaseq.BowtieBuild, 'bowtie-build',
                                         rnaseq.Bowtie, 'bowtie')

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2'),
                     'bowtie2 is not in the PATH')
    def test_StepBowtie2_noread(self):
        self._test_StepAlign_noreads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                     rnaseq.Bowtie2, 'bowtie2')
    @unittest.skipIf(not environment.Executable.ispresent('bowtie2'),
                     'bowtie2 is not in the PATH')
    def test_StepBowtie2_singlereads(self):
        self._test_StepAlign_singlereads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                         rnaseq.Bowtie2, 'bowtie2')
    @unittest.skipIf(not environment.Executable.ispresent('bowtie2'),
                     'bowtie2 is not in the PATH')
    def test_StepBowtie2_pairedreads(self):
        self._test_StepAlign_pairedreads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                         rnaseq.Bowtie2, 'bowtie2')


    @unittest.skipIf(not environment.Executable.ispresent('tophat'),
                     'tophat is not in the PATH')
    def test_StepTopHat_noread(self):
        self._test_StepAlign_noreads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                     rnaseq.TopHat, 'tophat')
    @unittest.skipIf(not environment.Executable.ispresent('tophat'),
                     'tophat is not in the PATH')
    def test_StepTopHat_singlereads(self):
        self._test_StepAlign_singlereads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                         rnaseq.TopHat, 'tophat')
    @unittest.skipIf(not environment.Executable.ispresent('tophat'),
                     'tophat is not in the PATH')
    def test_StepTopHat_pairedreads(self):
        self._test_StepAlign_pairedreads(rnaseq.BowtieBuild, 'bowtie2-build',
                                         rnaseq.TopHat, 'tophat')

    @unittest.skipIf(not environment.Executable.ispresent('tophat2'),
                     'tophat is not in the PATH')
    def test_StepTopHat2_noread(self):
        self._test_StepAlign_noreads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                     rnaseq.TopHat2, 'tophat2')
    @unittest.skipIf(not environment.Executable.ispresent('tophat2'),
                     'tophat2 is not in the PATH')
    def test_StepTopHat2_singlereads(self):
        self._test_StepAlign_singlereads(rnaseq.Bowtie2Build, 'bowtie2-build',
                                         rnaseq.TopHat2, 'tophat2')
    @unittest.skipIf(not environment.Executable.ispresent('tophat2'),
                     'tophat2 is not in the PATH')
    def test_StepTopHat2_pairedreads(self):
        self._test_StepAlign_pairedreads(rnaseq.BowtieBuild, 'bowtie2-build',
                                         rnaseq.TopHat, 'tophat2')

    @unittest.skipIf(not environment.Executable.ispresent('gsnap'),
                     'gsnap is not in the PATH')
    @unittest.skipIf(not environment.Executable.ispresent('gmap_build'),
                     'gmap_build is not in the PATH')
    def test_StepGsnapAlign_singlereads(self):
        self._test_StepAlign_singlereads(rnaseq.GsnapIndex, 'gmap_build',
                                         rnaseq.GsnapAlign, 'gsnap',
                                         parameters_index = ('-k', '8'))
    @unittest.skipIf(not environment.Executable.ispresent('gsnap'),
                     'gsnap is not in the PATH')
    @unittest.skipIf(not environment.Executable.ispresent('gmap_build'),
                     'gmap_build is not in the PATH')
    def test_StepGsnapAlign_pairedreads(self):
        self._test_StepAlign_pairedreads(rnaseq.GsnapIndex, 'gmap_build',
                                         rnaseq.GsnapAlign, 'gsnap',
                                         parameters_index = ('-k', '8'))

    @unittest.skipIf(not environment.Executable.ispresent('STAR'),
                     'STAR is not in the PATH')
    def test_StepStarAlign_singlereads(self):
        # STAR is whimsical. Try running the tests up to 5 times.
        self._test_StepAlign_singlereads(rnaseq.StarIndex, 'STAR',
                                         rnaseq.StarAlign, 'STAR',
                                         parameters_index = ('--genomeChrBinNbits', '12',
                                                             '--genomeSAindexNbases', '8'),
                                         align_max_attempts = 3)

    @unittest.skipIf(not environment.Executable.ispresent('STAR'),
                     'STAR is not in the PATH')
    def test_StepStarAlign_pairedreads(self):
        # STAR is whimsical. Try running the tests up to 5 times.
        self._test_StepAlign_pairedreads(rnaseq.StarIndex, 'STAR',
                                         rnaseq.StarAlign, 'STAR',
                                         parameters_index = ('--genomeChrBinNbits', '12',
                                                             '--genomeSAindexNbases', '8'),
                                         align_max_attempts = 3)


class ModelHandleBAMTestCase(unittest.TestCase):
    """
    Tests for the handling of BAM files (Steps and associated).
    """

    def setUp(self):
        NFRAGMENTS_NOMATCH = 10
        cls_index = rnaseq.Bowtie2Build
        executable_index = 'bowtie2-build'
        cls_align = rnaseq.Bowtie2
        executable_align = 'bowtie2'
        executable_samtools = 'samtools'
        read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq')
        read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq')
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh, reference)

        # reads from a random reference (should not match)
        randint = railroadtracks.model.simulate.random.randint
        rand_reference = railroadtracks.model.simulate.Entry('> Random DNA',
                                                             bytearray(''.join('ATCG'[randint(0, 3)] for x in range(500)), 'ascii'))
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh,
                                                                                     rand_reference,
                                                                                     n = NFRAGMENTS_NOMATCH)

        self.tempdir = tempfile.mkdtemp()
        assets_align, cmd, returncode, fh = _build_UpToAlign(cls_index, executable_index,
                                                             cls_align, executable_align, self._read1_fh,
                                                             os.path.join(self.tempdir, 'reference'))
        self._assets_align = assets_align

    def tearDown(self):
        self._read1_fh.close()
        self._read2_fh.close()
        shutil.rmtree(self.tempdir)


    @unittest.skipIf(not (environment.Executable.ispresent('samtools') and \
                          environment.Executable.ispresent('bowtie2-build')),
                     'samtools and bowtie2-build should be in the PATH')        
    def test_SamtoolsSortByID(self):
        SamtoolsSorterByID = rnaseq.SamtoolsSorterByID
        AssetsSorter = SamtoolsSorterByID.Assets
        executable_sort = 'samtools'
        samsort_byID = SamtoolsSorterByID(executable_sort)
        self.assertEqual(set((rnaseq.ACTIVITY.SORT,)), set(samsort_byID.activities))
        # mostly to check that version is working
        self.assertTrue(isinstance(samsort_byID.version, bytes))
        source = AssetsSorter.Source(self._assets_align.target.alignment)
        fh = tempfile.NamedTemporaryFile(suffix='.bam')
        sortedreads = rnaseq.BAMFile(fh.name)
        target = AssetsSorter.Target(sortedreads)
        assets = AssetsSorter(source, target)
        cmd, returncode = samsort_byID.run(assets)
        self.assertEqual(0, returncode)
        # check that the sorted BAM is not an empty file
        self.assertGreater(os.stat(fh.name).st_size, 0)

    @unittest.skipIf(not (environment.Executable.ispresent('samtools') and \
                          environment.Executable.ispresent('bowtie2-build')),
                     'samtools and bowtie2-build should be in the PATH')        
    def test_SamtoolsExtractUnaligned(self):
        SamtoolsExtractUnaligned = rnaseq.SamtoolsExtractUnaligned
        executable_sort = 'samtools'
        extractunaligned = rnaseq.SamtoolsExtractUnaligned(executable_sort)
        self.assertEqual(set((rnaseq.ACTIVITY.UTILITY,)), set(extractunaligned.activities))
        # mostly to check that version is working
        self.assertTrue(isinstance(extractunaligned.version, bytes))
        Assets = SamtoolsExtractUnaligned.Assets
        source = Assets.Source(self._assets_align.target.alignment)
        fh = tempfile.NamedTemporaryFile(suffix='.bam')
        sortedreads = rnaseq.BAMFile(fh.name)
        target = Assets.Target(sortedreads)
        assets = Assets(source, target)
        cmd, returncode = extractunaligned.run(assets)
        self.assertEqual(0, returncode)
        #FIXME: other tests on the results


    @unittest.skipIf(not (environment.Executable.ispresent('samtools') and \
                          environment.Executable.ispresent('bowtie2-build')),
                     'samtools and bowtie2-build should be in the PATH')        
    def test_SamtoolsFilter(self):
        SamtoolsFilter = railroadtracks.model.files.SamtoolsFilter
        executable_sort = 'samtools'
        readfilter = SamtoolsFilter(executable_sort)
        self.assertEqual(set((railroadtracks.model.files.ACTIVITY.FILTERREADS,)), set(readfilter.activities))
        # mostly to check that version is working
        self.assertTrue(isinstance(readfilter.version, bytes))
        Assets = SamtoolsFilter.Assets
        source = Assets.Source(self._assets_align.target.alignment)
        # 
        fh_f0x4 = tempfile.NamedTemporaryFile(suffix='.bam')
        sortedreads = rnaseq.BAMFile(fh_f0x4.name)
        target = Assets.Target(sortedreads)
        assets = Assets(source, target)
        cmd, returncode = readfilter.run(assets, ('--filter-include', '0x0004'))
        self.assertEqual(0, returncode)
        in_stat = os.stat(assets.source.bamfile.name)
        out_stat_inc = os.stat(assets.target.bamfile.name)
        self.assertGreater(in_stat.st_size, out_stat_inc.st_size)
        #
        fh_F0x4 = tempfile.NamedTemporaryFile(suffix='.bam')
        sortedreads = rnaseq.BAMFile(fh_F0x4.name)
        target = Assets.Target(sortedreads)
        assets = Assets(source, target)
        cmd, returncode = readfilter.run(assets, ('--filter-exclude', '0x0004'))
        self.assertEqual(0, returncode)
        out_stat_exc = os.stat(assets.target.bamfile.name)
        self.assertGreater(out_stat_exc.st_size, out_stat_inc.st_size)


class ModelAlignQuantificationTestCase(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq', dir=self.tempdir)
        read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq', dir=self.tempdir)
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
            # set the random seed
            random.seed(a=123)
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh, reference)

    def tearDown(self):
        self._read1_fh.close()
        self._read2_fh.close()
        shutil.rmtree(self.tempdir)

    @unittest.skipIf(not environment.Executable.ispresent('sailfish'),
                     'sailfish is not in the PATH')        
    def test_pairedend_Sailfish(self):
        execname = 'sailfish'
        assets, cmd, returncode = _build_StepIndex(railroadtracks.model.aligners.SailfishIndex(execname),
                                                   execname, 
                                                   os.path.join(self.tempdir, 'reference'),
                                                   parameters = runner.PARAMETERS_DEFAULT)
        if returncode != 0:
            raise Exception("Index not built")
        self.assets = assets
        sf = railroadtracks.model.quantify.SailfishQuant(execname)
        A = sf.Assets
        count_csv = os.path.join(self.tempdir, 'counts.csv')
        output_dir = os.path.join(self.tempdir, 'sailfish_output_dir')
        assets = A(A.Source(self.assets.target.indexfilepattern,
                            rnaseq.FASTQPossiblyGzipCompressed(self._read1_fh.name),
                            rnaseq.FASTQPossiblyGzipCompressed(self._read2_fh.name)),
                   A.Target(rnaseq.CSVFile(count_csv),
                            rnaseq.FilePattern(output_dir)))
        cmd, returncode = sf.run(assets, parameters = ('--libtype', sf.LIBRARY_PE))

    @unittest.skipIf(not environment.Executable.ispresent('salmon'),
                     'salmon is not in the PATH')
    def test_pairedend_Salmon(self):
        execname = 'salmon'
        SalmonIndex = railroadtracks.model.aligners.SalmonIndex
        assets, cmd, returncode = _build_StepIndex(SalmonIndex, execname,
                                                   os.path.join(self.tempdir, 'reference'),
                                                   parameters = ('--threads', '1'))
        if returncode != 0:
            raise Exception("Index not built")
        self.assets = assets
        sm = railroadtracks.model.quantify.SalmonAlignQuant(execname)
        A = sm.Assets
        count_csv = os.path.join(self.tempdir, 'counts.csv')
        output_dir = os.path.join(self.tempdir, 'salmon_output_dir')
        assets = A(A.Source(self.assets.target.indexfilepattern,
                            rnaseq.FASTQPossiblyGzipCompressed(self._read1_fh.name),
                            rnaseq.FASTQPossiblyGzipCompressed(self._read2_fh.name)),
                   A.Target(rnaseq.CSVFile(count_csv),
                            rnaseq.FilePattern(output_dir)))
        cmd, returncode = sm.run(assets, parameters = ('--libType', sm.LIBRARY_PE))

class ModelQuantificationTestCase(unittest.TestCase):
    """
    """

    def setUp(self):
        cls_index = rnaseq.Bowtie2Build
        executable_index = 'bowtie2-build'
        cls_align = rnaseq.Bowtie2
        executable_align = 'bowtie2'
        self.tempdir = tempfile.mkdtemp()
        read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq')
        read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq')
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
        self._read1_fh, self._read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh, reference)
        assets_align, cmd, returncode, fh = _build_UpToAlign(cls_index, executable_index,
                                                             cls_align, executable_align,
                                                             self._read1_fh,
                                                             os.path.join(self.tempdir, 'reference'))
        self._assets_align = assets_align
        self._fh = fh # protect temp file from deletion

    def tearDown(self):
        self._read1_fh.close()
        self._read2_fh.close()
        shutil.rmtree(self.tempdir)

    @unittest.skipIf(not (environment.Executable.ispresent('htseq-count') and \
                          environment.Executable.ispresent('bowtie2-build')),
                     'htseq-count is not in the PATH')        
    def test_HTSeqCount(self):
        AssetsQuantifier = railroadtracks.model.quantify.AssetsQuantifier
        HTSeqCount = rnaseq.HTSeqCount
        referenceannotation = rnaseq.GFFFile(PHAGEGFF)
        executable_count = 'htseq-count'
        # non-default parameters to fit our demo GFF
        htseqcount = HTSeqCount(executable_count)
        self.assertEqual(set((rnaseq.ACTIVITY.QUANTIFY,)), set(htseqcount.activities))

        version = htseqcount.version

        assets_align = self._assets_align
        # sort to remove the warning
        stsort = rnaseq.SamtoolsSorterByID()
        with tempfile.NamedTemporaryFile(suffix='.bam') as sortedbam_fh, \
             tempfile.NamedTemporaryFile(suffix='.csv') as fh:
            assets_sort = stsort.Assets(stsort.Assets.Source(assets_align.target.alignment),
                                        stsort.Assets.Target(railroadtracks.model.files.BAMFileSortedByID(sortedbam_fh.name)))
            stsort.run(assets_sort)
            source = AssetsQuantifier.Source(assets_sort.target.sortedbam,
                                             referenceannotation)
            
            counts = rnaseq.CSVFile(fh.name)
            target = AssetsQuantifier.Target(counts)
            assets = AssetsQuantifier(source, target)
            cmd, returncode = htseqcount.run(assets,
                                             parameters=HTSeqCount._noexons_parameters)
            self.assertEqual(0, returncode)
            cmd, returncode = htseqcount.run(assets,
                                             parameters=itertools.chain(HTSeqCount._noexons_parameters,
                                                                        ('-m', 'union', '-s', 'no', '-a', '10')))
            self.assertEqual(0, returncode)

    @unittest.skipIf(not (environment.Executable.ispresent('htseq-count') and \
                          environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('Rsubread') is not None),
                     'Htseq-count is not in the PATH')        
    def test_FeatureCount(self):
        AssetsQuantifier = railroadtracks.model.quantify.AssetsQuantifier
        FeatureCount = rnaseq.FeatureCount
        referenceannotation = rnaseq.GFFFile(PHAGEGFF)
        # wrapped R script
        executable_count = 'R'
        # non-default parameters to fit our demo GFF
        featurecount = FeatureCount(executable_count)
        self.assertEqual(set((rnaseq.ACTIVITY.QUANTIFY,)), set(featurecount.activities))

        version = featurecount.version

        assets_align = self._assets_align
        source = AssetsQuantifier.Source(assets_align.target.alignment,
                                         referenceannotation)
        fh = tempfile.NamedTemporaryFile(suffix='.csv')
        counts = rnaseq.CSVFile(fh.name)
        target = AssetsQuantifier.Target(counts)
        assets = AssetsQuantifier(source, target)
        cmd, returncode = featurecount.run(assets, 
                                           parameters=featurecount._noexons_parameters)
        self.assertEqual(0, returncode)
        # try with paired-end (the total of counts should now be approx. half)
        cmd, returncode = featurecount.run(assets, parameters=tuple(itertools.chain(featurecount._noexons_parameters, featurecount._pe_parameters)))
        self.assertEqual(0, returncode)

class ModelColumnMergerTestCase(unittest.TestCase):

    def test_merge(self):
        with tempfile.NamedTemporaryFile(mode='w+', suffix=".csv") as fh_col1, \
             tempfile.NamedTemporaryFile(mode='w+', suffix=".csv") as fh_col2, \
             tempfile.NamedTemporaryFile(mode='w+', suffix=".csv") as fh_merged:
            fh_col1_csv = csv.writer(fh_col1)
            fh_col2_csv = csv.writer(fh_col2)
            for x in ('1','2','3'):
                fh_col1_csv.writerow([x])
            fh_col1.seek(0)
            for x in ('4','5','6'):
                fh_col2_csv.writerow([x])
            fh_col2.seek(0)
            ColumnMerger = rnaseq.ColumnMerger
            Assets = ColumnMerger.Assets
            merger = ColumnMerger()
            filestomerge = tuple(rnaseq.CSVFile(x.name) for x in (fh_col1, fh_col2))
            source = Assets.Source(rnaseq.CSVFileSequence(filestomerge))
            assets = Assets(source,
                            Assets.Target(rnaseq.CSVFile(fh_merged.name)))
            cmd, returncode = merger.run(assets,
                                         parameters=('None', '0'))
            self.assertEqual(0, returncode)
            # just in case... rewind to 0
            fh_merged.seek(0)
            self.assertEqual(('1,4',
                              '2,5',
                              '3,6'), tuple(x.rstrip() for x in fh_merged))

    def test_merge_withIDs(self):
        with tempfile.NamedTemporaryFile(suffix=".csv", mode='w+') as fh_col1, \
             tempfile.NamedTemporaryFile(suffix=".csv", mode='w+') as fh_col2, \
             tempfile.NamedTemporaryFile(suffix=".csv", mode='w+') as fh_merged:
            fh_col1_csv = csv.writer(fh_col1)
            fh_col2_csv = csv.writer(fh_col2)
            for row in (('1','a'),('2','b'),('3','c')):
                fh_col1_csv.writerow(row)
            fh_col1.flush()
            fh_col1.seek(0)
            for row in (('1','d'),('2','e'),('3','f')):
                fh_col2_csv.writerow(row)
            fh_col2.flush()
            fh_col2.seek(0)
            ColumnMerger = rnaseq.ColumnMerger
            Assets = ColumnMerger.Assets
            merger = ColumnMerger()
            filestomerge = tuple(rnaseq.CSVFile(x.name) for x in (fh_col1, fh_col2))
            source = Assets.Source(rnaseq.CSVFileSequence(filestomerge))
            assets = Assets(source,
                            Assets.Target(rnaseq.CSVFile(fh_merged.name)))
            cmd, returncode = merger.run(assets, parameters=('0', '1'))
            self.assertEqual(0, returncode)
            # just in case... rewind to 0
            fh_merged.seek(0)
            self.assertEqual(('1,a,d',
                              '2,b,e',
                              '3,c,f'),
                             tuple(x.rstrip() for x in fh_merged))

class ModelDExpressionTestCase(unittest.TestCase):
    """
    """
    
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        sampleinfo_fh = tempfile.NamedTemporaryFile(dir=self.tempdir, 
                                                    mode='w+',
                                                    suffix='.csv', 
                                                    delete=False)
        csv_w = csv.writer(sampleinfo_fh)
        csv_w.writerow(['sample_id', 'group'])
        for i in range(6):
            csv_w.writerow([str(i), ('A','B')[i%2]])
        sampleinfo_fh.flush()
        self._sampleinfo_fh = sampleinfo_fh
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
        random.seed(123)    
        with tempfile.NamedTemporaryFile(dir=self.tempdir,
                                         mode='w+',
                                         suffix='.csv', 
                                         delete=False) as fh_merged:
            csv_w = csv.writer(fh_merged)
            self.asset_merge = rnaseq.CSVFile(fh_merged.name)
            self._r_exec = environment.R('R')
            csv_w.writerow(['',] + ['V%i'%x for x in range(1,7)])
            with open(PHAGEGFF) as gff_fh:
                csv_r = csv.reader(gff_fh, delimiter="\t")
                for row in csv_r:
                    if row[0].startswith('#'):
                        continue
                    if row[2] != 'CDS':
                        continue
                    entry_id = re.sub('ID="(.+?)".+', '\\1', row[8])
                    count_row = [entry_id, ] + [int(random.lognormvariate(4, 2)) for x in range(6)]
                    csv_w.writerow(count_row)
            
                    
    def tearDown(self):
        shutil.rmtree(self.tempdir)


    @unittest.skipIf(not (environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('DESeq') is not None),
                     'R (with package "DESeq") must be in the PATH')
    def test_StepDESeq(self):
        deseq = rnaseq.DESeq(self._r_exec)
        self.assertEqual(set((rnaseq.ACTIVITY.DIFFEXP,)), set(deseq.activities))
        AssetsDE = railroadtracks.model.diffexp.AssetsDifferentialExpression
        fh = tempfile.NamedTemporaryFile(mode='w+')
        source = AssetsDE.Source(self.asset_merge, 
                                 core.File(self._sampleinfo_fh.name))
        target = AssetsDE.Target(core.File(fh.name))
        assets = AssetsDE(source,
                          target)
        cmd, returncode = deseq.run(assets, parameters=('--dispersion-fittype=local', ))
        self.assertEqual(0, returncode)

    @unittest.skipIf(not (environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('DESeq2') is not None),
                     'R (with package "DESeq2") must be in the PATH')
    def test_StepDESeq2(self):
        deseq2 = rnaseq.DESeq2(self._r_exec)
        self.assertEqual(set((rnaseq.ACTIVITY.DIFFEXP,)), set(deseq2.activities))
        AssetsDE = railroadtracks.model.diffexp.AssetsDifferentialExpression
        fh = tempfile.NamedTemporaryFile(mode='w+')
        source = AssetsDE.Source(self.asset_merge,
                                 core.File(self._sampleinfo_fh.name))
        target = AssetsDE.Target(core.File(fh.name))
        assets = AssetsDE(source,
                          target)
        cmd, returncode = deseq2.run(assets)
        self.assertEqual(0, returncode)


    @unittest.skipIf(not (environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('DESeq2') is not None),
                     'R (with package "edgeR") must be in the PATH')
    def test_StepEdgeR(self):
        edger = rnaseq.EdgeR(self._r_exec)
        self.assertEqual(set((rnaseq.ACTIVITY.DIFFEXP,)), set(edger.activities))
        AssetsDE = railroadtracks.model.diffexp.AssetsDifferentialExpression
        fh = tempfile.NamedTemporaryFile(mode='w+')
        source = AssetsDE.Source(self.asset_merge,
                                 core.File(self._sampleinfo_fh.name))
        target = AssetsDE.Target(core.File(fh.name))
        assets = AssetsDE(source,
                          target)
        cmd, returncode = edger.run(assets)
        self.assertEqual(0, returncode)


    @unittest.skipIf(not (environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('limma') is not None),
                     'R (with package "limma") must be in the PATH')
    def test_StepLimmaVoom(self):
        voom = rnaseq.LimmaVoom(self._r_exec)
        self.assertEqual(set((rnaseq.ACTIVITY.DIFFEXP,)), set(voom.activities))
        AssetsDE = railroadtracks.model.diffexp.AssetsDifferentialExpression
        fh = tempfile.NamedTemporaryFile(mode='w+')
        source = AssetsDE.Source(self.asset_merge,
                                 core.File(self._sampleinfo_fh.name))
        target = AssetsDE.Target(core.File(fh.name))
        assets = AssetsDE(source,
                          target)
        cmd, returncode = voom.run(assets)
        self.assertEqual(0, returncode)


    @unittest.skipIf(not (environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('EBSeq') is not None),
                     'R (with package "EBSeq") must be in the PATH')
    def test_StepEBSeq(self):
        ebseq = railroadtracks.model.diffexp.EBSeq(self._r_exec)
        self.assertEqual(set((railroadtracks.model.diffexp.ACTIVITY.DIFFEXP,)), set(ebseq.activities))
        AssetsDE = railroadtracks.model.diffexp.AssetsDifferentialExpression
        fh = tempfile.NamedTemporaryFile(mode='w+')
        source = AssetsDE.Source(self.asset_merge,
                                 core.File(self._sampleinfo_fh.name))
        target = AssetsDE.Target(core.File(fh.name))
        assets = AssetsDE(source,
                          target)
        cmd, returncode = ebseq.run(assets)
        self.assertEqual(0, returncode)


class ModelCRCHeadTailTestCase(unittest.TestCase):

    def _test(self, data, crc):
        fh = tempfile.NamedTemporaryFile(mode='w+')
        fh.write(data)
        fh.flush()
        out_fh = tempfile.NamedTemporaryFile(mode='w+', suffix='.csv')
        assets = crc.Assets(crc.Assets.Source(core.File(fh.name)),
                            crc.Assets.Target(rnaseq.CSVFile(out_fh.name)))
        cmd, returncode = crc.run(assets)
        self.assertEqual(0, returncode)
        csv_r = csv.reader(out_fh)
        header = next(csv_r)
        filename, crc = next(csv_r)
        self.assertEqual(fh.name, filename)
        return crc

    def test_init(self):
        CRC = rnaseq.CRCHeadTail
        crc = CRC(None)
        data = 'foobarbaz'
        crc_a = self._test(data, crc)
        #
        crc_b = self._test(data, crc)
        #
        self.assertEqual(crc_a, crc_b)
        #
        data = '123456789'
        crc_c = self._test(data, crc)
        self.assertNotEqual(crc_a, crc_c)


class GzipFastqFilePairTestCase(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq.gz', dir=self.tempdir, delete=False)
        read1_fh.close()
        self.read1_fn = read1_fh.name

        read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq.gz', dir=self.tempdir, delete=False)
        read2_fh.close()
        self.read2_fn = read2_fh.name

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @unittest.skipIf(not has_ngsp,
                     'The Python package ngs-plumbing is missing.')
    def test_GzipFastqFilePair(self):
        NFRAGMENTS_MATCH = 300
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
            read1_io = gzip.GzipFile(self.read1_fn, mode='w')
            read2_io = gzip.GzipFile(self.read2_fn, mode='w')
            # reads from the target genome
            read1_fh, read2_fh=railroadtracks.model.simulate.randomPEreads(read1_io,
                                                                           read2_io,
                                                                           reference,
                                                                           n = NFRAGMENTS_MATCH)
        read1_fh.close()
        read2_fh.close()
        fqp = rnaseq.GzipFastqFilePair(self.read1_fn,
                                       self.read2_fn)
        readpairs = tuple(fqp)
        for i, (r1,r2) in enumerate(readpairs):
            self.assertTrue(hasattr(r1, 'header'))
            self.assertTrue(hasattr(r1, 'sequence'))
            self.assertTrue(hasattr(r1, 'quality'))
            self.assertTrue(hasattr(r2, 'header'))
            self.assertTrue(hasattr(r2, 'sequence'))
            self.assertTrue(hasattr(r2, 'quality'))
        self.assertEqual(NFRAGMENTS_MATCH, i+1)
        
if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest, tempfile, csv, shutil, logging

from railroadtracks import core, hortator, rnaseq, easy, environment

from railroadtracks.model.simulate import (PHAGEGFF,
                                           PHAGEGTF,
                                           PHAGEFASTA)
import railroadtracks.model.simulate

from railroadtracks.test.test_model import (_build_UpToAlign,
                                            _build_StepIndex,
                                            _build_StepAlign,
                                            _build_StepQuantify)


# Test the writing of recipes
from railroadtracks import easy

class RecipeTestCase(unittest.TestCase):

    def setUp(self):
        # -- recipe-init-begin
        # -- initialization boiler plate code
        wd = tempfile.mkdtemp()
        project = easy.Project(rnaseq, wd=wd)

        # declare the 3rd-party command-line tools we will use
        env = easy.Environment(rnaseq)
        # -- recipe-init-end

        # -- recipe-data-begin
        # Phage genome shipped with the package for testing purposes

        PHAGEFASTA = railroadtracks.model.simulate.PHAGEFASTA
        PHAGEGFF = railroadtracks.model.simulate.PHAGEGFF

        # create random data for 6 samples (just testing here)
        nsamples = 6
        samplereads = list()
        with open(PHAGEFASTA) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))
        for sample_i in range(nsamples):
            read1_fh = tempfile.NamedTemporaryFile(prefix='read1', suffix='.fq')
            read2_fh = tempfile.NamedTemporaryFile(prefix='read2', suffix='.fq')
            read1_fh, read2_fh = railroadtracks.model.simulate.randomPEreads(read1_fh,
                                                                             read2_fh, 
                                                                             reference)
            samplereads.append((read1_fh, read2_fh))

        sampleinfo_fh = tempfile.NamedTemporaryFile(suffix='.csv', mode='w+')
        csv_w = csv.writer(sampleinfo_fh)
        csv_w.writerow(['sample_id', 'group'])
        for i in range(6):
            csv_w.writerow([str(i), ('A','B')[i%2]])
        sampleinfo_fh.flush()
        referenceannotation = rnaseq.GFFFile(PHAGEGFF)
        # -- recipe-data-end

        self._wd = wd
        self.project = project
        self.reference_fn = PHAGEFASTA
        self.env = env
        self.nsamples = nsamples
        self.samplereads = samplereads
        self.sampleinfo_fh = sampleinfo_fh
        self.referenceannotation = referenceannotation
        self._PHAGEFASTA = PHAGEFASTA
        self._PHAGEGFF = PHAGEGFF

    def tearDown(self):
        samplereads = self.samplereads
        # -- recipe-teardown-begin
        for read1_fh, read2_fh in self.samplereads:
            read1_fh.close()
            read2_fh.close()
        # FIXME: delete the temporary directory
        shutil.rmtree(self.project.wd)
        # -- recipe-teardown-end

    def test_File(self):
        #FIXME: rather test it in the model ?
        reference = core.File(self.reference_fn)

    @unittest.skipIf(not (environment.Executable.ispresent('bowtie2-build') and \
                          environment.Executable.ispresent('htseq-count') and \
                          environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('edgeR') is not None),
                     'bowtie2 and/or htseq-count is not in the PATH')
    def test_RecipeSimpleIncremental(self):
        project = self.project
        env = self.env
        nsamples = self.nsamples
        samplereads = self.samplereads
        sampleinfo_fh = self.sampleinfo_fh
        reference_fn = self.reference_fn
        referenceannotation = self.referenceannotation
        PHAGEFASTA = self._PHAGEFASTA
        PHAGEGFF = self._PHAGEGFF

        # steps used
        bowtie2index = env.activities.INDEX.bowtie2build
        bowtie2align = env.activities.ALIGN.bowtie2
        htseqcount = env.activities.QUANTIFY.htseqcount
        merge = env.activities.UTILITY.columnmerger
        edger = env.activities.DIFFEXP.edger

        from railroadtracks import easy

        # sequence of tasks to run
        torun = list()

        # index for alignment
        Assets = bowtie2index.Assets
        assets = Assets(Assets.Source(rnaseq.FASTAFile(reference_fn)),
                        Assets.Target.createundefined())
        task_index = project.add_task(bowtie2index, 
                                      assets)
        # the step is not done
        self.assertEqual(hortator._TASK_TODO, task_index.info[1])
        torun.append(task_index)
        # run the tasks
        for task in torun:
            # run only if not done
            if task.info[1] != hortator._TASK_DONE:
                task.execute()
                task.status = hortator._TASK_DONE

        self.assertEqual(1, project.persistent_graph.nconcrete_steps)
        # now that the tasks have run let's open the same project
        project_same = easy.Project(project.model, wd=project.wd)

        # index for alignment
        Assets = bowtie2index.Assets
        assets = Assets(Assets.Source(rnaseq.FASTAFile(reference_fn)),
                        Assets.Target.createundefined())
        task_index_same = project_same.add_task(bowtie2index, 
                                                assets)

        self.assertNotEqual(task_index, task_index_same)
        self.assertNotEqual(task_index.call.assets, task_index_same.call.assets)
        self.assertListEqual(list(task_index.call.assets.source.reference), 
                             list(task_index_same.call.assets.source.reference))
        self.assertListEqual(list(task_index.call.assets.target.indexfilepattern), 
                             list(task_index_same.call.assets.target.indexfilepattern))
        self.assertEqual(hortator._TASK_DONE, task_index_same.info[1])
        self.assertEqual(1, project.persistent_graph.nconcrete_steps)


    def _recipesimpleincremental(self, runtasks):
        project = self.project
        env = self.env
        nsamples = self.nsamples
        samplereads = self.samplereads
        sampleinfo_fh = self.sampleinfo_fh
        reference_fn = self.reference_fn
        referenceannotation = self.referenceannotation
        PHAGEFASTA = self._PHAGEFASTA
        PHAGEGFF = self._PHAGEGFF
        
        # steps used
        bowtie2index = env.activities.INDEX.bowtie2build
        bowtie2align = env.activities.ALIGN.bowtie2
        htseqcount = env.activities.QUANTIFY.htseqcount
        merge = env.activities.UTILITY.columnmerger
        edger = env.activities.DIFFEXP.edger

        for iteration in range(5):
            nextiteration = False
            # sequence of tasks to run
            torun = list()

            # index for alignment
            Assets = bowtie2index.Assets
            assets = Assets(Assets.Source(rnaseq.FASTAFile(reference_fn)),
                            Assets.Target.createundefined())
            task_index = project.add_task(bowtie2index, assets)
            torun.append(task_index)
            if iteration < 1:
                nextiteration = True
                runtasks(torun)
                self.assertEqual(1, project.persistent_graph.nconcrete_steps)
                continue
            # process all samples
            sample_counts = list()
            for sample_i, (read1_fh, read2_fh) in enumerate(samplereads):
                # align
                Assets = bowtie2align.Assets
                assets = Assets(Assets.Source(task_index.call.assets.target.indexfilepattern, 
                                              rnaseq.FASTQPossiblyGzipCompressed(read1_fh.name),
                                              rnaseq.FASTQPossiblyGzipCompressed(read2_fh.name)),
                                Assets.Target.createundefined())
                task_align = project.add_task(bowtie2align, assets)
                torun.append(task_align)
                if iteration < 2:
                    nextiteration = True
                    runtasks(torun)
                    self.assertEqual(1+(sample_i+1), project.persistent_graph.nconcrete_steps)
                    continue

                # quantify
                # (non-default parameters to fit our demo GFF)
                params = rnaseq.HTSeqCount._noexons_parameters
                Assets = htseqcount.Assets
                assets = Assets(Assets.Source(task_align.call.assets.target.alignment,
                                              rnaseq.GFFFile(referenceannotation)),
                                Assets.Target.createundefined())
                task_quantify = project.add_task(htseqcount,
                                                 assets,
                                                 parameters=params)
                torun.append(task_quantify)
                if iteration < 3:
                    nextiteration = True
                    runtasks(torun)
                    self.assertEqual(1+len(samplereads)+(sample_i+1), 
                                     project.persistent_graph.nconcrete_steps)
                    continue

                # keep a pointer to the counts, as we will use it in the merge step
                sample_counts.append(task_quantify.call.assets)

            if nextiteration:
                continue
            # merge the sample data into a table (so differential expression can be computed)
            Assets = merge.Assets
            counts = tuple(x.target.counts for x in sample_counts)
            assets = Assets(Assets.Source(rnaseq.CSVFileSequence(counts)),
                            merge.Assets.Target.createundefined())

            task_merge = project.add_task(merge,
                                          assets,
                                          parameters=("0", "1"))
            torun.append(task_merge)
            if iteration < 4:
                nextiteration = True
                runtasks(torun)
                self.assertEqual(1+2*len(samplereads)+1, 
                                 project.persistent_graph.nconcrete_steps)
                continue

            # differential expression with edgeR
            Assets = edger.Assets
            assets = Assets(Assets.Source(task_merge.call.assets.target.counts,
                                          rnaseq.CSVFile(sampleinfo_fh.name)),
                            Assets.Target.createundefined())
            task_de = project.add_task(edger,
                                       assets)
            if iteration < 5:
                nextiteration = True
                runtasks(torun)
                self.assertEqual(1+2*len(samplereads)+2, # 1 index + 2 FASTQ per sample + 1 merge + 1 differential expression
                                 project.persistent_graph.nconcrete_steps)
                continue

    @unittest.skipIf(not (environment.Executable.ispresent('bowtie2-build') and \
                          environment.Executable.ispresent('htseq-count') and \
                          environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('edgeR') is not None),
                     'bowtie2 and/or htseq-count is not in the PATH')
    def test_RecipeSimpleIncrementalComplete(self):
        def runtasks(torun):
            # run the tasks
            for task in torun:
                # run only if not done
                if task.info[1] != hortator._TASK_DONE:
                    task.execute()
        self._recipesimpleincremental(runtasks)

    @unittest.skipIf(not (environment.Executable.ispresent('bowtie2-build') and \
                          environment.Executable.ispresent('htseq-count') and \
                          environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('edgeR') is not None),
                     'bowtie2, htseq-count, R (with package "edgeR") must be in the PATH')
    def test_RecipeSimpleIncrementalCompleteNoRun(self):
        def runtasks(torun):
            # do nothing
            pass
        self._recipesimpleincremental(runtasks)


    @unittest.skipIf(not (environment.Executable.ispresent('bowtie2-build') and \
                          environment.Executable.ispresent('htseq-count') and \
                          environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('edgeR') is not None),
                     'bowtie2, htseq-count, R (with package "edgeR") must be in the PATH')
    def test_RecipeSimple(self):
        project = self.project
        env = self.env
        nsamples = self.nsamples
        samplereads = self.samplereads
        sampleinfo_fh = self.sampleinfo_fh
        reference_fn = self.reference_fn
        referenceannotation = self.referenceannotation
        PHAGEFASTA = self._PHAGEFASTA
        PHAGEGFF = self._PHAGEGFF
        
        # -- recipesimple-test-begin

        # steps used
        bowtie2index = env.activities.INDEX.bowtie2build
        bowtie2align = env.activities.ALIGN.bowtie2
        htseqcount = env.activities.QUANTIFY.htseqcount
        merge = env.activities.UTILITY.columnmerger
        edger = env.activities.DIFFEXP.edger

        from railroadtracks import easy

        # sequence of tasks to run
        torun = list()
                            
        # index for alignment
        Assets = bowtie2index.Assets
        assets = Assets(Assets.Source(rnaseq.FASTAFile(reference_fn)),
                        Assets.Target.createundefined())
        task_index = project.add_task(bowtie2index, assets)
        torun.append(task_index)

        # process all samples
        sample_counts = list()
        for read1_fh, read2_fh in samplereads:
            # align
            Assets = bowtie2align.Assets
            assets = Assets(Assets.Source(task_index.call.assets.target.indexfilepattern, 
                                          rnaseq.FASTQPossiblyGzipCompressed(read1_fh.name),
                                          rnaseq.FASTQPossiblyGzipCompressed(read2_fh.name)),
                            Assets.Target.createundefined())
            task_align = project.add_task(bowtie2align, assets)
            torun.append(task_align)

            # quantify
            # (non-default parameters to fit our demo GFF)
            params = rnaseq.HTSeqCount._noexons_parameters
            Assets = htseqcount.Assets
            assets = Assets(Assets.Source(task_align.call.assets.target.alignment,
                                          rnaseq.GFFFile(referenceannotation)),
                            Assets.Target.createundefined())
            task_quantify = project.add_task(htseqcount,
                                             assets,
                                             parameters=params)
            torun.append(task_quantify)
            # keep a pointer to the counts,
            # as we will use them in the merge step
            sample_counts.append(task_quantify.call.assets)

        # merge the sample data into a table
        # (so differential expression can be computed)
        Assets = merge.Assets
        counts = tuple(x.target.counts for x in sample_counts)
        assets = Assets(Assets.Source(rnaseq.CSVFileSequence(counts)),
                        merge.Assets.Target.createundefined())
        task_merge = project.add_task(merge,
                                      assets,
                                      parameters=("0","1"))
        torun.append(task_merge)

        # differential expression with edgeR
        Assets = edger.Assets
        assets = Assets(Assets.Source(task_merge.call.assets.target.counts,
                                      rnaseq.CSVFile(sampleinfo_fh.name)),
                        Assets.Target.createundefined())
        task_de = project.add_task(edger,
                                   assets)

        # run the tasks
        for task in torun:
            # run only if not done
            if task.info[1] != hortator._TASK_DONE:
                task.execute()

        # get results
        final_storedentities = project.get_targetsofactivity(rnaseq.ACTIVITY.DIFFEXP)

        # get the step that created the results files
        final_steps = list()
        for stored_entity in final_storedentities:
            final_steps.append(project.persistent_graph.get_parenttask_of_storedentity(stored_entity))
        
        # -- recipesimple-test-end
        
        self.assertEqual(1, len(final_storedentities))
        self.assertEqual(core.File.__name__, final_storedentities[0].clsname)
        self.assertEqual('railroadtracks.model.diffexp.EdgeR', final_steps[0].clsname)

        # FIXME: not yet implemented
        # now that we have all steps, we "only" have to run them
        #steps = todo.stepcrawler()
        #for s in steps:
        #    print('%s' % (s.unifiedname))
        #    s.run()

    @unittest.skipIf(not (environment.Executable.ispresent('bowtie2-build') and \
                          environment.Executable.ispresent('bowtie-build') and \
                          environment.Executable.ispresent('STAR') and \
                          environment.Executable.ispresent('R') and \
                          environment.R('R').packageversion_or_none('edgeR') is not None and \
                          environment.R('R').packageversion_or_none('DESeq') is not None and \
                          environment.R('R').packageversion_or_none('DESeq2') is not None and \
                          environment.R('R').packageversion_or_none('limma') is not None),
                     'bowtie2, bowtie, STAR, TopHat2, and R (with packages "edgeR", "DESeq", "DESeq2", "limma") must be in the PATH')
    def test_RecipeLoop(self):
        project = self.project
        env = self.env
        nsamples = self.nsamples
        samplereads = self.samplereads
        sampleinfo_fh = self.sampleinfo_fh
        reference_fn = self.reference_fn
        referenceannotation = self.referenceannotation
        PHAGEFASTA = self._PHAGEFASTA
        PHAGEGFF = self._PHAGEGFF

        # -- recipeloop-test-begin
        from railroadtracks import easy

        torun = list()

        # bowtie
        bowtie1index = env.activities.INDEX.bowtiebuild
        bowtie1align = env.activities.ALIGN.bowtie
        Assets = bowtie1index.Assets
        fa_file = rnaseq.FASTAFile(reference_fn)
        task_index_bowtie1 = project.add_task(bowtie1index, 
                                              Assets(Assets.Source(fa_file),
                                                     None))
        torun.append(task_index_bowtie1)

        # bowtie2
        bowtie2index = env.activities.INDEX.bowtie2build
        bowtie2align = env.activities.ALIGN.bowtie2
        Assets = bowtie2index.Assets
        fa_file = rnaseq.FASTAFile(reference_fn)
        task_index_bowtie2 = project.add_task(bowtie2index,
                                              Assets(Assets.Source(fa_file),
                                                     None))
        torun.append(task_index_bowtie2)

        # STAR
        starindex = env.activities.INDEX.starindex
        staralign = env.activities.ALIGN.staralign
        Assets = starindex.Assets
        fa_file = rnaseq.FASTAFile(reference_fn)
        task_index_star = project.add_task(starindex, 
                                           Assets(Assets.Source(fa_file),
                                                  None))
        torun.append(task_index_star)

        # TopHat2
        # (index from bowtie2 used)
        #tophat2 = env.activities.ALIGN.tophat2

        # featureCount
        featurecount = env.activities.QUANTIFY.featurecount

        # Merge columns (obtained from counting)
        merge = env.activities.UTILITY.columnmerger

        # EdgeR, DESeq, DESeq2, and LIMMA voom
        edger = env.activities.DIFFEXP.edger
        deseq = env.activities.DIFFEXP.deseq
        deseq2 = env.activities.DIFFEXP.deseq2
        voom = env.activities.DIFFEXP.limmavoom
        

        # Now explore the different alignment presets in bowtie2, and vanilla star
        from itertools import cycle
        from collections import namedtuple
        Options = namedtuple('Options', 'aligner assets_index parameters')
        # Try various presets for bowtie2
        bowtie2_parameters = (('--very-fast', ), ('--fast', ), 
                              ('--sensitive', ), ('--very-sensitive', ))
        options = [Options(*x) for x in zip(cycle((bowtie2align,)),
                                            cycle((task_index_bowtie2.call.assets.target,)),
                                            bowtie2_parameters)]

        # add bowtie
        options.append(Options(bowtie1align, task_index_bowtie1.call.assets.target, tuple()))
        # add STAR (vanilla, no specific options beside the size of index k-mers)
        options.append(Options(staralign, 
                               task_index_star.call.assets.target, 
                               ('--genomeChrBinNbits', '12')))
        # add TopHat2
        #options.append(Options(tophat2, task_index_bowtie2.call.assets.target, tuple()))

        # loop over the options
        for option in options:
            sample_counts = list()
            # loop over the samples
            for sample_i in range(nsamples):
                read1_fh, read2_fh = samplereads[sample_i]
                # align
                Assets = option.aligner.Assets
                assets = Assets(Assets.Source(option.assets_index.indexfilepattern,
                                              rnaseq.FASTQPossiblyGzipCompressed(read1_fh.name), 
                                              rnaseq.FASTQPossiblyGzipCompressed(read2_fh.name)),
                                Assets.Target.createundefined())
                task_align = project.add_task(option.aligner,
                                              assets,
                                              parameters=option.parameters)
                torun.append(task_align)

                # quantify
                # (non-default parameters to fit our demo GFF)
                Assets = featurecount.Assets
                assets = Assets(Assets.Source(task_align.call.assets.target.alignment,
                                              rnaseq.GFFFile(referenceannotation)),
                                Assets.Target.createundefined())
                task_quantify = project.add_task(featurecount,
                                                 assets,
                                                 parameters = ('--gtf-featuretype', 'CDS',
                                                               '--gtf-attrtype', 'ID'))
                torun.append(task_quantify)

                # keep a pointer to the counts, as we will use it in the merge step
                sample_counts.append(task_quantify.call.assets)

            # merge the sample data into a table (so differential expression can be computed)
            Assets = merge.Assets
            source = Assets.Source(rnaseq.CSVFileSequence(tuple(x.target.counts\
                                                                for x in sample_counts)))
            assets_merge = Assets(source,
                                  Assets.Target.createundefined())
            task_merge = project.add_task(merge,
                                          assets_merge,
                                          parameters=("0","1"))
            torun.append(task_merge)

            # differential expression with edgeR, deseq2, and voom
            # (deseq is too whimsical for tests)
            for diffexp, params in ((edger, ()),
                                    (deseq, ('--dispersion-fittype=local', )), 
                                    (deseq2, ()),
                                    (voom, ())):
                Assets = diffexp.Assets
                assets = Assets(Assets.Source(task_merge.call.assets.target.counts,
                                              core.File(sampleinfo_fh.name)),
                                Assets.Target.createundefined())
                task_de = project.add_task(diffexp,assets)
                torun.append(task_de)

        # run the tasks
        # (this is an integration test rather than a unit test - the 
        # 3rd-party tools are often brittle and we want to keep the noise level down)
        env_log_level = environment.logger.level
        environment.logger.level = logging.ERROR
        try:
            for task in torun:
                if task.info[1] != hortator._TASK_DONE:
                    try:
                        task.execute()
                        status = easy.hortator._TASK_DONE
                    except:
                        status = easy.hortator._TASK_FAILED
                project.persistent_graph.step_concrete_state(hortator.DbID(task.task_id, False),
                                                             easy.hortator._TASK_STATUS_LIST[status])
        finally:
            environment.logger.level = env_log_level
        # -- recipeloop-test-end


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unified execution layer.

One general way to run things on the command line.
"""

import sys, os, argparse
import collections
import logging
logger = logging.getLogger(__name__)
import railroadtracks
from importlib import import_module
from . import environment
from . import core


LOG_DEBUG = 'DEBUG'
LOGGING_LEVELS = ('INFO', LOG_DEBUG)

class UnifexError(Exception):
    pass



#FIXME: several classes 'Call' in rrt, I think. Delete/rename others
class Call(object):
    """ Unified call, turning a step + assets + parameters into a task. """
    def __init__(self, step, assets, parameters):
        self._step = step
        self._assets = assets
        self._parameters = parameters

    def execute(self):
        """ Execute the task. """
        return self._step.run(self._assets, self._parameters)

    @property
    def step(self):
        return self._step
    @property
    def assets(self):
        return self._assets
    @property
    def parameters(self):
        return self._parameters

    def __str__(self):
        res = (super(Call, self).__str__(),
               '  step: %s' % str(self.step),
               '  assets: %s' % str(self.assets),
               '  parameters: %s' % str(self.parameters))
        return os.linesep.join(res)

def _set_logging(args):
    if args.logging_file is not None:
        logging.basicConfig(filename=args.logging_file,
                            level=getattr(logging, args.logging_level))

def _make_stepdict(module):
    """ 
    Make a dict of classes (key=step name, value=class).
    Duplicate step names will raise a ValueError.
    :param classlist: module 
    :rtype: :class:`namespace`
    """
    classlist = core.steplist(module)
    d = dict()
    for cls in classlist:
        if not issubclass(cls, (core.StepAbstract,)):
            raise ValueError("Classes used as step must inherit from core.StepAbstract.")
        stepname = cls._name
        if stepname is None:
            raise ValueError('The step name for class "%s" is not defined.' % cls)
        elif stepname in d:
            raise ValueError('The step name "%s" is defined twice.' % stepname)
        d[stepname] = cls
    return d


def _extract_argdict(arglist):
    """ Build dictionary from list of parameters.
    """
    argdict = collections.defaultdict(list)
    if arglist is not None:
         for src in arglist:
            key, value = src.split('=', 1)
            argdict[key].append(value)
    return argdict

def _extract_arglist(argdict):
    """ Build list of parameters from a mapping of parameters.
    """
    arglist = list()
    if argdict is not None:
        for key, value in argdict.items():
            for v in value:
                src = '='.join((key, v))
                arglist.append(src)
    return tuple(arglist)


#FIXME: moved to unifex.py ?
def _cmdfromuei(uei):
    cmd = [sys.executable, '-m', 'railroadtracks.unifex', 'run', uei.model]
    # In the unified command line, the executable does not have to be specified,
    # and when parsing command line arguments with argparse the associated
    # parameter is set to None.
    # When building a unified command line, we skip the executable if it is
    # set to None.
    if uei.executable is not None:
        cmd.append(uei.executable)
    cmd.append('-s')
    cmd.extend('%s=%s' % (x,y.name) for x,y in zip(uei.source._fields, uei.source))
    cmd.append('-t')
    cmd.extend('%s=%s' % (x,y.name) for x,y in zip(uei.target._fields, uei.target))
    if len(uei.parameters) > 0:
        cmd.append('-p')
        cmd.extend(" '%s'" % x for x in uei.parameters)
    return cmd


#FIXME: needed ?
# def wrapper_run(model_cls, sources, targets,
#                 parameters):
#     executable = model_cls(args.executable, parameters = args.parameters)
#     res = executable.run(sources, targets)
#     return res


def _model_instance(args, steplist):
    """
    :param args: arguments, such as the ones returned by :meth:`argparse.ArgumentParser.parse`.
    
    This should contain 2 attributes:
    - :attr:`model`, the name of the model used for execution,
    - :attr:`executable` the name (or path) to the executable used.

    :param steplist: sequence of known steps
    :rtype: :class:``
    """
    model_cls = steplist[args.model]
    try:
        model_instance = model_cls(args.executable)
    except environment.MissingSoftware:
        raise UnifexError("""The executable '%s' to use with the model class '%s' is not in the ${PATH}.
Use either the full path, or add the executable to your PATH.
        """ % (args.executable, args.model))
    except Exception as e:
        raise UnifexError("""Internal error while creating a %s using %s:
%s
""" % (model_cls,
       args.executable,
       e))
    return model_instance


def build_AssetSet(AssetSet, 
                   values):
    """
    :param AssetSet:
    :param values: values to create instances in the AssetSet
    """
    # classes expected for the assets
    assetset_cls = getattr(AssetSet, core.AssetMetaReserved.SOURCES.value)
    assets = list()
    for x, val in zip(assetset_cls, values):
        if val is None:
            # if None, whether it is allowed (allownone True/False) is pushed to the construction
            # of the AssetSet below
            #FIXME: is this really the most transparent way to do it ?
            a = val
        else:
            if issubclass(x.cls, core.FileSequence):
                a = x.cls(x.cls._type(z) for z in val)
            else:
                a = x.cls(val)
        assets.append(a)
    assetset = AssetSet(*assets)
    return assetset

def unified_exec_run(args, steplist, msg=[]):
    """
    Run a command.
    :param args: arguments in a class such as the one returned by :meth:`argparse.ArgumentParser.parse`
    :param steplist: sequence of known steps. The `args` will be matched against this to find the model class.
    :type steplist: sequence of :class:`core.StepAbstract`-inherting instances
    :param msg: list with (eventual) messages
    """
    #FIXME: deprecate msg (use logging)
    _set_logging(args)
    model = _model_instance(args, steplist)
    Assets = model.Assets

    # extract the sources from the args
    sources_dict = _extract_argdict(args.source)
    tmp = ['Source parameters:', ]
    tmp.extend('%s: %s%s' % (k, str(v), os.linesep) for k,v in sources_dict.items())
    logger.debug(os.linesep.join(tmp))
    sources = tuple(sources_dict.get(f) for f in Assets.Source._fields)
    sources_undefined = list()
    for src, assetattr, assetname in zip(sources, Assets.Source._sources, Assets.Source._fields):
        if src is None:
            if not assetattr.allownone:
                sources_undefined.append(assetname)
    if len(sources_undefined) > 0:
        msg.append('The following sources must be defined (and are missing):\n'+\
                   '\n'.join('- %s' % sources_undefined))

    # extract the targets from the args
    targets_dict = _extract_argdict(args.target)
    targets = tuple(targets_dict.get(f) for f in Assets.Target._fields)

    if None in targets:
        msg.append('The following targets must be defined (and are missing):\n'+\
                   '\n'.join('- %s' % (y) for x,y in enumerate(Assets.Target._fields) if targets[x] is None))

    # exit early if missing parameters
    if len(msg) > 0:
        raise ValueError('\n'.join(msg))

    # build the assets
    assets = Assets(build_AssetSet(Assets.Source, sources),
                    build_AssetSet(Assets.Target, targets))
    cleanparameters = tuple(x.lstrip() for x in args.parameters)
    cmd, returncode = model.run(assets, cleanparameters)

    msg = 'return code: %i' % returncode
    logger.debug(msg)
    if returncode != 0:
        sys.stderr.write(msg)
        sys.stderr.write('\n')
    return returncode

def unified_exec_version(args, steplist):
    _set_logging(args)
    executable = _model_instance(args, steplist)
    print(executable.version)
    return 0

def unified_exec_activities(args, steplist):
    _set_logging(args)
    executable = _model_instance(args, steplist)
    print(executable.activities)
    return 0

def unified_exec_model(args, steplist):
    _set_logging(args)
    print(' '.join(str(x) for x in steplist.keys()))
    return 0

def unified_exec():

    def _steplist(module_name):
        module = import_module(module_name)
        steplist = _make_stepdict(module)
        return steplist

    def _unified_exec_run(args):
        msg = []
        try:
            returncode = unified_exec_run(args, _steplist(args.module), msg=msg)
        except ValueError as ve:
            logger.error(msg)
            raise ve
        return returncode

    def _unified_exec_version(args):
        return unified_exec_version(args, _steplist(args.module))
    def _unified_exec_activities(args):
        return unified_exec_activities(args, _steplist(args.module))
    def _unified_exec_model(args):
        return unified_exec_model(args, _steplist(args.module))

    parser = argparse.ArgumentParser(description='Unified model for RNA-Seq steps')
    subparsers = parser.add_subparsers(help='Action to perform. `run` will call the executable specified in the next parameter, `version` will return the version number, `activities` will list the RNA-Seq activities the executable is associated with')

    #FIXME: This means that one will be able to point to different R version, and packages will be coming
    #       from that R version


    # run
    parser_run = subparsers.add_parser('run', help="Run a step")
    #FIXME: This means that one will be able to point to different R version, and packages will be coming
    #       from that R version
    parser_run.add_argument('model',
                            help='Class name used to model that executable. The classes are defined in the model (see -m/--module).')
    parser_run.add_argument('executable',
                            nargs='?', # this is optional
                            help='Name, or full path, for an executable. '
                            'Note that the executable can be R, '
                            'and in that case the class name (argument "model") '
                            'will wrap an R script to be run with that R version.')
    parser_run.add_argument('-m', '--module',
                            default='railroadtracks.rnaseq',
                            help='module defining all models (default: %(default)s)')
    parser_run.add_argument('-s', '--source', nargs='*',
                            help='Source file indicated as <label>=<filename>')
    parser_run.add_argument('-t', '--target', nargs='*',
                            help='target file indicated as <label>=<filename>')
    parser_run.add_argument('-p', '--parameters', nargs='*',
                            default = (),
                            help='Additional parameter(s) for the wrapped executable')
    parser_run.add_argument('--logging-file',
                            help='Name of log file.')
    parser_run.add_argument('--logging-level',
                            choices = LOGGING_LEVELS,
                            default = 'INFO',
                            help='Level of logging information. DEBUG is quite verbose (default: %(default)s).')
    parser_run.set_defaults(func = _unified_exec_run)

    # version
    parser_version = subparsers.add_parser('version',
                                           help="Query version number")
    parser_version.add_argument('model',
                                help='Class name used to model that executable. The classes are defined in the model (see -m/--module).')
    parser_version.add_argument('executable', 
                                nargs='?', # this is optional
                                help='Name, or full path, for an executable. '
                                'Note that the executable can be R, '
                                'and in that case the class name (argument "model") '
                                'will wrap an R script to be run with that R version.')
    parser_version.add_argument('-m', '--module',
                                default='railroadtracks.rnaseq',
                                help='module defining all models (default: %(default)s)')
    parser_version.add_argument('--logging-file',
                                help='Name of log file.')
    parser_version.add_argument('--logging-level',
                                choices = ('INFO', 'DEBUG'),
                                default = 'INFO',
                                help='Level of logging information. DEBUG is quite verbose (default: %(default)s).')

    parser_version.set_defaults(func = _unified_exec_version)

    # activities
    parser_activities = subparsers.add_parser('activities',
                                              help="Query activities")
    parser_activities.add_argument('model',
                                   help='Class name used to model that executable. The classes are defined in the model (see -m/--module).')
    parser_activities.add_argument('-m', '--module',
                                   default='railroadtracks.rnaseq',
                                   help='module defining all models (default: %(default)s)')
    parser_activities.add_argument('executable',
                                   nargs='?', # this is optional
                                   help='Name, or full path, for an executable. '
                                   'Note that the executable can be R, '
                                   'and in that case the class name (argument "model") '
                                   'will wrap an R script to be run with that R version.')
    parser_activities.add_argument('--logging-file',
                                   help='Name of log file.')
    parser_activities.add_argument('--logging-level',
                                   choices = ('INFO', 'DEBUG'),
                                   default = 'INFO',
                                   help='Level of logging information. DEBUG is quite verbose (default: %(default)s).')

    parser_activities.set_defaults(func = _unified_exec_activities)

    # model
    parser_model = subparsers.add_parser('model',
                                         help="Query on a model module")
    parser_model.add_argument('-m', '--module',
                              default='railroadtracks.rnaseq',
                              help='module defining all models (default: %(default)s)')
    parser_model.add_argument('--logging-file',
                              help='Name of log file.')
    parser_model.add_argument('--logging-level',
                              choices = ('INFO', 'DEBUG'),
                              default = 'INFO',
                              help='Level of logging information. DEBUG is quite verbose (default: %(default)s).')

    parser_model.set_defaults(func = _unified_exec_model)
    
    args = parser.parse_args()
    returncode = args.func(args)


if __name__ == '__main__':
    import sys
    returncode = unified_exec()
    sys.exit(returncode)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import division
import os
from railroadtracks import core, easy
from railroadtracks.hortator import _TASK_DONE
from collections import namedtuple
from jinja2 import Environment, PackageLoader
from railroadtracks import core, hortator, easy
import railroadtracks.easy.taskgraph
import railroadtracks.easy.tasksetgraph
import networkx
try:
    import pygraphviz
    has_pygraphviz = True
except:
    has_pygraphviz = False

env = Environment(loader=PackageLoader('railroadtracks'),
                  extensions = ['jinja2.ext.with_'])



def format_largenumber(value):
    return "{:,.2f}".format(value)
env.filters['format_largenumber'] = format_largenumber


def html_project_view_storage(project, cssclass='rsrsproject'):
    d = easy.dict_project_view_storage(project)
    d['cssclass'] = cssclass
    template = env.get_template('project_view_storage.html')
    res = template.render(d)
    return res
    

def html_project_view_activities(project, done=_TASK_DONE,
                                 cssclass='rsrsproject'):
    template = env.get_template('project_view_activities.html')
    d = easy.dict_project_view_activities(project, done=done)
    d['cssclass'] = cssclass
    res = template.render(d)
    return res


def html_project_view_results(project,
                              cssclass='rsrsproject'):
    d = easy.dict_project_view_results(project)
    d['cssclass'] = cssclass
    template = env.get_template('project_view_results.html')
    res = template.render(d)
    return res



rsrsproject_css = """
table.rsrsproject {
  width: 100%;
  max-width: 40em;
  border-collapse:separate;
  border:solid black 2px;
  -moz-border-radius:6px;
  -webkit-border-radius:6px;
  border-radius:6px;
}
table.rsrsproject thead {
  background-color:rgb(235,235,235);
}
table.rsrsproject th:first-child {
  -moz-border-radius:6px 0 0 0;
  -webkit-border-radius:6px 0 0 0;
  border-radius:6px 0 0 0;
}
table.rsrsproject th:last-child {
  -moz-border-radius: 0 6px 0 0;
  -webkit-border-radius: 0 6px 0 0;
  border-radius: 0 6px 0 0;
}
table.rsrsproject th:only-child {
  -moz-border-radius: 6px 6px 0 0;
  -webkit-border-radius: 6px 6px 0 0;
  border-radius: 6px 6px 0 0;
}
table.rsrsproject .inverse {
  background-color: rgb(50,50,50);
  color: rgb(230,230,230);
}

progress.rsrsproject {
  -webkit-appearance:none;
  -moz-appearance:none;
  appearance:none;
}
progress[value].rsrsproject::-webkit-progress-bar {
  background-color: rgb(230,230,230);
  box-shadow: 0 2px 5px rgba(0,0,0,0.25) inset;
}
progress[value].rsrsproject::-webkit-progress-value {
  background-color: rgb(150,150,250);
}
progress[value].rsrsproject {
  border: none;
  box-shadow: 0 2px 5px rgba(0,0,0,0.25) inset;
}
progress[value].rsrsproject::-moz-progress-bar {
  background-color: rgb(150,150,250);
}
"""

def html_project_view(project,
                      cssclass='rsrsproject'):
    res = ['<style>', rsrsproject_css, '</style>']
    res.extend((x(project, cssclass=cssclass) for x in (html_project_view_storage,
                                                        html_project_view_activities,
                                                        html_project_view_results)))
    return os.linesep.join(res)


def html_assetset_view(assetset,
                       cssclass='rsrsproject'):
    d = easy.dict_assetset_view(assetset)
    d['cssclass'] = cssclass    
    template = env.get_template('assetset_view.html')
    res = template.render(d)
    return res


def html_task_view(task,
                   cssclass='rsrsproject'):
    d = dict()
    d['source_elements'] = easy.dict_assetset_view(task.call.assets.source)['elements']
    d['target_elements'] = easy.dict_assetset_view(task.call.assets.target)['elements']
    d['parameters'] = task.call.parameters
    d['task'] = task
    d['task_class'] = type(task.call.step).__name__
    d['cssclass'] = cssclass    
    template = env.get_template('task_view.html')
    res = template.render(d)
    return res

def node_label(instance):
    if isinstance(instance, hortator.StoredEntity):
        label_ext = "."+instance.entityname.split(".")[-1]
        label_cls = instance.clsname
        if len(label_ext) < len(label_cls):
            label = label_ext
        else:
            label = label_cls
        label = 't:%i - %s' % (instance.id, label)
    elif isinstance(instance, hortator.StoredSequence):
        label_cls = instance.clsname
        label = label_cls
        label = 's:%i - %s' % (instance.id, label)
    else:
        label = 'a:%i - %s' % (instance.task_id, instance.call.step._name)
    return label

def basic_tasknode_func(ag, inst, nodelabel):
    node = ag.get_node(nodelabel)
    node.attr['shape'] = 'box'
    node.attr['tooltip'] = 'parameters: %s' % ' '.join(str(x) for x in inst.call.parameters)
    return node

def default_tasknode_func(ag, inst, nodelabel):
    node = basic_tasknode_func(ag, inst, nodelabel)
    if inst.status == easy._TASK_FAILED:
        node.attr.update(style='filled', fillcolor='black', fontcolor="white")
    elif inst.status == easy._TASK_DONE:
        node.attr.update(style='filled', fillcolor='white', fontcolor="black", penwidth=1.5)
    else:
        node.attr.update(style='filled', fillcolor='white', color="grey", fontcolor="grey")
    return node

def default_assetnode(ag, inst, nodelabel, font_size, pred_done):
    node = ag.get_node(nodelabel)
    node.attr['fontsize'] = int(font_size*.75)
    node.attr['tooltip'] = inst.entityname
    if not pred_done:
        node.attr['fontcolor'] = "grey"
        node.attr['color'] = "grey"
    return node

def default_assetseqnode(ag, inst, nodelabel, font_size, pred_done):
    node = ag.get_node(nodelabel)
    node.attr['fontsize'] = int(font_size*.75)
    if not pred_done:
        node.attr['fontcolor'] = "grey"
        node.attr['color'] = "grey"
    return node

def _get_pred_done(node, digraph):
    pred_done = True
    for pnode in digraph.predecessors(node):
        instance2 = digraph.node[pnode]['instance']
        if not hasattr(instance2, 'status'):
            pred_done = _get_pred_done(pnode, digraph)
        elif instance2.status != hortator._TASK_DONE:
            pred_done = False
            break
    return pred_done


def agraph_fromdigraph(digraph, layout = "dot",
                       nodelabel_func = node_label,
                       tasknode_func = default_tasknode_func,
                       assetnode_func = default_assetnode,
                       assetseqnode_func = default_assetseqnode,
                       graph_size = None,
                       font_size = 14):
    """
    Plot a :class:`DiGraph` from networkx.

    :param digraph: a :class:`DiGraph`
    :param format: the output format
    :param layout: the layout algorithm (from graphviz)
    :param nodelabel_func:
    :param tasknode_func:
    :param assetnode_func:
    :param assetseqnode_func:
    :param graph_size: attribute "size" for the graph (e.g., "3,2")
    :param font_size: font size
    """
    if not has_pygraphviz:
        raise RuntimeError('The Python package "pygraphviz" is required but could not be imported.')
    ag = pygraphviz.AGraph(directed=True)
    if graph_size is not None:
        ag.graph_attr.update(size = graph_size)
    ag.graph_attr.update(font_size = font_size)
    for node in networkx.topological_sort(digraph):
        instance = digraph.node[node]['instance']
        nodelabel = nodelabel_func(instance)
        for node2 in digraph.successors(node):
            instance2 = digraph.node[node2]['instance']
            if _get_pred_done(node2, digraph):
                # if all predecessors are done, default edge display
                ag.add_edge(nodelabel, nodelabel_func(instance2))
            else:
                # if not all predecessors done, lighter edge
                ag.add_edge(nodelabel, nodelabel_func(instance2), color="grey")
        if isinstance(instance, easy.Task):
            tasknode_func(ag, instance, nodelabel)
            pred_done = instance.status == hortator._TASK_DONE
        else:
            pred_done = _get_pred_done(node, digraph)
            if isinstance(instance, hortator.StoredSequence):
                assetseqnode_func(ag, instance, nodelabel, font_size, pred_done)
            else:
                assetnode_func(ag, instance, nodelabel, font_size, pred_done)
    ag.layout(prog=layout)
    return ag


def _plot_digraph(digraph, format, graph_size=None):
    ag = agraph_fromdigraph(digraph, graph_size=graph_size)
    string = ag.draw(format=format)
    return string

def svg_digraph(digraph, graph_size=None):
    return _plot_digraph(digraph, 'svg', graph_size=graph_size)

def svg_taskgraph_view(taskgraph, graph_size=None):
    svg = svg_digraph(taskgraph.digraph)
    return svg

def agraph_fromtasksetgraph(tasksetgraph, taskgraph,
                            graph_size=None,
                            nodelabel_func = node_label, layout="dot"):
    ag = agraph_fromdigraph(taskgraph.digraph, graph_size=graph_size, layout=layout)
    el = tasksetgraph.execution_list()
    knownlabels = dict()
    for tasksetbundle in el:
        if len(tasksetbundle.taskset) == 1:
            continue
        nodebundle = list()
        label = tasksetbundle.label
        if label is None:
            a_set = set()
            for task in tasksetbundle.taskset:
                a_set.update(set(a.value for a in task.activities)),
            label = '/'.join(a_set)

        if label not in knownlabels:
            knownlabels[label] = 1
        else:
            knownlabels[label] += 1
            label = "%s - %i" % (label, knownlabels[label])

        for task in tasksetbundle.taskset:
            #node = easy.taskgraph.Node('Task', task.task_id)
            nodelabel = nodelabel_func(task)
            nodebundle.append(nodelabel)

        ag.add_subgraph(nodebundle, name='cluster_' + label, label=label, color="grey")
    # likely bug in pygraphviz - this is needed
    ag.layout(prog=layout)
    return ag
        
def svg_tasksetgraph_view(tasksetgraph, taskgraph,
                          graph_size=None,
                          nodelabel_func = node_label, layout="dot"):
    ag = agraph_fromtasksetgraph(tasksetgraph, taskgraph,
                                 graph_size=graph_size,
                                 nodelabel_func = node_label, layout="dot")
    svg = ag.draw(format='svg')
    return svg


def _project_status_curses(stdscr, project, checkinterval=3):
    import time, curses, datetime
    time_lastupdate = os.path.getmtime(project._db_fn)
    while True:
        pos_x = 0
        pos_y = 0
        stdscr.addstr(pos_x, pos_y, 'Fetching status...', curses.A_REVERSE)
        stdscr.refresh()
        str_storage = easy.str_project_view_storage(project, width=15)
        str_activities = easy.str_project_view_activities(project, width=15)
        str_results = easy.str_project_view_results(project, width=15)        
        for row in str_storage.split(os.linesep):
            stdscr.addstr(pos_x, pos_y, row)
            pos_x += 1
        stdscr.addstr(pos_x, pos_y, '    Activities    ', curses.A_REVERSE)
        pos_x += 1
        for row in str_activities.split(os.linesep):
            stdscr.addstr(pos_x, pos_y, row)
            pos_x += 1
        stdscr.addstr(pos_x, pos_y, '    Results    ', curses.A_REVERSE)
        pos_x += 1
        for row in str_results.split(os.linesep):
            stdscr.addstr(pos_x, pos_y, row)
            pos_x += 1

        pos_x += 2
        pos_update_x = pos_x
        tdelta = datetime.timedelta(0, 
                                    int(time.time()-time_lastupdate), # seconds 
                                    0, #microseconds 
                                    0, #milliseconds
                                    0)
        stdscr.addstr(pos_update_x, pos_y, '    [Last status change was %s ago]  ' % tdelta,
                      curses.A_DIM)
        pos_x += 1
        stdscr.addstr(pos_x, pos_y, 
                      '    [Checking updates every %.1f seconds]' % (checkinterval), 
                      curses.A_DIM)
        pos_x += 1
        stdscr.addstr(pos_x, pos_y, '    [Ctrl-c to exit]  ',
                      curses.A_DIM)
        stdscr.refresh()
        time_lastupdate = os.path.getmtime(project._db_fn)
        while time_lastupdate == os.path.getmtime(project._db_fn):
            time.sleep(checkinterval)
            tdelta = datetime.timedelta(0, 
                                        int(time.time()-time_lastupdate), # seconds 
                                        0, #microseconds 
                                        0, #milliseconds
                                        0)
            stdscr.addstr(pos_update_x, pos_y, '    [Last status change was %s ago]  ' % tdelta,
                          curses.A_DIM)
            stdscr.refresh()

def _project_status_str(args):
    import importlib
    model = importlib.import_module(args.model)
    project = easy.Project(model, args.wd, db_fn=args.db_fn)
    if args.follow:
        import curses
        import sys
        try:
            curses.wrapper(_project_status_curses, project, checkinterval=args.checkinterval)
        except KeyboardInterrupt:
            sys.exit(1)
    else:
        print(easy.str_project_view(project))


# Custom display
def init_printing():
    from railroadtracks import easy, core
    ip = get_ipython()
    html_f = ip.display_formatter.formatters['text/html']
    html_f.for_type(easy.Project, html_project_view)
    html_f.for_type(core.AssetSet, html_assetset_view)
    html_f.for_type(easy.Task, html_task_view)
    svg_f = ip.display_formatter.formatters['image/svg+xml']
    svg_f.for_type(easy.taskgraph.TaskGraph, svg_taskgraph_view)

if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='rsrs project tools')
    subparsers = parser.add_subparsers(help='Action to perform. `status` will report the status of the project.')

    parser_status = subparsers.add_parser('status', help="Report status")

    parser_status.add_argument('wd',
                               help='Working directory')
    parser_status.add_argument('model',
                               help='Model (as a Python module)')
    parser_status.add_argument('--db-fn', dest='db_fn',
                               default = None,
                               help='File name for the database')
    parser_status.add_argument('-f', '--follow', dest='follow',
                               action='store_true',
                               help='Follow the status')
    parser_status.add_argument('-t', '--time',
                               dest='checkinterval',
                               type=float,
                               default=3,
                               help='Time interval to update status (in seconds)')
    parser_status.set_defaults(func = _project_status_str)
    
    args = parser.parse_args()
    args.func(args)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" 
Very minimal set of functions to simulate NGS data.

There are much more advanced tools available. This is mostly here to run unit-tests and
related activities.

"""

import railroadtracks
import warnings
import argparse
import logging
logger = logging.getLogger(__name__)
from railroadtracks import environment, core, unifex
from railroadtracks.model.files import (File, 
                                        FilePattern,
                                        GFFFile,
                                        FASTQPossiblyGzipCompressed,
                                        BEDFile)
import sys, os, argparse, tempfile, re
import random
import collections
from collections import deque
import csv
import subprocess
import shutil

if sys.version_info[0] < 3:
    linesep = os.linesep
else:
    linesep = bytes(os.linesep, 'ascii')

PHAGEFASTA = os.path.join(os.path.dirname(railroadtracks.__file__), 'EF204940.fa')
PHAGEGFF = os.path.join(os.path.dirname(railroadtracks.__file__), 'ef204940.gff')
PHAGEGTF = os.path.join(os.path.dirname(railroadtracks.__file__), 'ef204940.gtf')

# Conditional definition for Python 3:
if sys.version_info[0] > 2:
    maketrans = bytearray.maketrans
else:
    from string import maketrans
COMPLEMENT_TABLE = maketrans(b'ATGC',b'TACG')

Entry = collections.namedtuple('Entry', 'header sequence')
def readfasta_iter(fh):
    """ Iterate through a FASTA file. """
    header = None
    sequence = None
    for row in fh:
        if row.startswith('>'):
            if header is not None:
                yield Entry(header, bytearray(''.join(sequence), 'ascii'))
            header = row.rstrip()
            sequence = list()
        else:
            sequence.append(row.strip())
    if len(sequence) > 0:
        yield Entry(header, bytearray(''.join(sequence), 'ascii'))

GFFEntry = collections.namedtuple('GFFEntry', 'seqname source feature start end score strand fram attribute')
GFFSTART_I = GFFEntry._fields.index('start')
GFFEND_I = GFFEntry._fields.index('end')
def readgff_iter(fh):
    """ Iterate through a GFF file 
    :param fh: iterable (for example a file object). """
    for row in fh:
        if row.startswith('#'):
            continue
        fields = row.split('\t')
        fields[GFFSTART_I] = int(fields[GFFSTART_I])
        fields[GFFEND_I] = int(fields[GFFEND_I])
        yield GFFEntry(*fields)

def entryfrom_gff(entry, gffentry):
    sequence = entry.sequence[gffentry.start:gffentry.end]
    if gffentry.strand == '+':
        pass
    else:
        sequence = sequence.translate(COMPLETEMENT_TABLE)
    return Entry(gffentry.seqname, sequence)

def randomreadstart(entry, length):
    """ random starting point in the sequence of an entry. """
    l = len(entry.sequence)
    if length > l:
        raise ValueError("Reads cannot be longer than the original reference.")
    return random.randint(0, l-length-1)

#FIXME: reverse-complement as well ?
READHEADER_TEMPLATE = '%(init)sFOO:%(lane)i:%(tile)i:%(x)i:%(y)i#0/%(pair)i'
def randomfastq(entry, n, length):
    dummyqual = b'~'*length
    for r_i in range(n):
        r_start = randomreadstart(entry, length)
        if sys.version_info[0] < 3:
            buf = buffer(entry.sequence, r_start, length)
        else:
            # Python 3
            buf = memoryview(entry.sequence)[r_start:(r_start+length)]
        res = (READHEADER_TEMPLATE % {'init':'@', 'x': r_i, 'y': r_i,
                                      'lane':1, 'tile':2, 'pair':1},
               buf,
               READHEADER_TEMPLATE % {'init':'+', 'x': r_i, 'y': r_i,
                                      'lane':1, 'tile':2, 'pair':1},
               dummyqual)
        yield res

if sys.version_info[0] < 3:
    def _readheader(init, r_i, lane, tile, pair):
        return bytes(READHEADER_TEMPLATE % {'init':init, 
                                            'x': r_i, 'y': r_i,
                                            'lane':lane, 'tile':tile, 
                                            'pair':pair})
else:
    def _readheader(init, r_i, lane, tile, pair):
        return bytes(READHEADER_TEMPLATE % {'init':init, 
                                            'x': r_i, 'y': r_i,
                                            'lane':lane, 'tile':tile, 
                                            'pair':pair}, 'ascii')

def randomfastq_pe(entry, n, length, insert, lane=1, tile=2):
    sequence_rc = entry.sequence[::-1].translate(COMPLEMENT_TABLE)
    quality_template = b'wxyz{|}~'
    dummyqual = (quality_template * ((length//len(quality_template))+1))[:length]

    for r_i in range(n):
        r_start = randomreadstart(entry, length+insert+length)
        if sys.version_info[0] < 3:
            buf = buffer(entry.sequence, r_start, length)
        else:
            # Python 3
            buf = memoryview(entry.sequence)[r_start:(r_start+length)]
        read1 = (_readheader('@', r_i, lane, tile, 1),
                 buf,
                 _readheader('+', r_i, lane, tile, 1),
                 dummyqual)
        if sys.version_info[0] < 3:
            buf = buffer(sequence_rc, 0, length)
        else:
            # Python 3
            buf = sequence_rc[:length]
        read2 = (_readheader('@', r_i, lane, tile, 2),
                 buf,
                 _readheader('+', r_i, lane, tile, 2),
                 dummyqual)
        yield (read1, read2)

def randomPEreads(read1_fh, read2_fh,
                  fastaentry, 
                  n=300, l=100, insert=50):
    """
    Random paired-end reads.
    :param read1_fh:
    :param read2_fh:
    :param fastaentry: FASTA entry
    :param n=200: number of reads
    :param l=100: length for the reads
    :return (read1_fh, read2_fh): pair of file handles 
    """
    # FIXME: rather minimalist (plan use of third-party generator of random sequencing
    # data)
    for read1, read2 in randomfastq_pe(fastaentry, n, l, insert):
        for row in read1:
            read1_fh.write(row)
            read1_fh.write(linesep)
        for row in read2:
            read2_fh.write(row)
            read2_fh.write(linesep)
    read1_fh.flush()
    read2_fh.flush()
    return (read1_fh, read2_fh)



# 

class ACTIVITY(core.Enum):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'SIMULATE'
    SIMULATE = 'Simulate'

def _split_mergedpairs(merged_fn, read1_fn, read2_fn):
    # split
    # (inspired by https://gist.github.com/nathanhaigh/3521724)

    inconsistent = False
    if read1_fn.endswith('.gz'):
        if read2_fn.endswith('.gz'):
            pipe_zip = '| gzip '
        else:
            inconsistent = True
    elif read2_fn.endswith('.gz'):
        inconsistent = True
    else:
        pipe_zip = ''
    if inconsistent:
        # this is presumably a mistake. do not allow it
        raise ValueError('The output files can be either both with the extention .gz, or both without.')
        
    env = os.environ.copy()
    env['mergedreads'] = subprocess.list2cmdline((merged_fn,))
    env['read1'] = subprocess.list2cmdline((read1_fn, ))
    env['read2'] = subprocess.list2cmdline((read2_fn,))
    cmd_str = 'paste - - - - - - - - < $mergedreads | ' + \
              'tee >(cut -f 1-4 | tr "\\t" "\\n" %(pipe_zip)s > $read1) | cut -f 5-8 | tr "\\t" "\\n" %(pipe_zip)s > $read2' % locals()
    with core.Popen(('/bin/bash', '-c', 
                     cmd_str),
                    env=env) as proc:
        proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError()


FLUXSIMULATOR_TEMPLATE = """
REF_FILE_NAME   %(annotation_gtf)s
GEN_DIR         %(genome_dir)s

NB_MOLECULES    %(nb_molecules)i

# error model
ERR_FILE        %(error)s

FRAG_SUBSTRATE  RNA
FRAG_METHOD	UR

READ_NUMBER	%(read_number)i
READ_LENGTH	%(read_length)i
PAIRED_END	true

# create a fastq file
FASTA           YES
"""
STATS_FILE_TEMPLATE = """
STATS_FILE   %(stats_fn)s
"""
TMP_DIR_TEMPLATE = """
TMP_DIR   %(tmp_dir)s
"""
LIB_FILE_TEMPLATE = """
LIB_FILE_NAME   %(lib_fn)s
"""
SEQ_FILE_TEMPLATE = """
SEQ_FILE_NAME   %(reads_fn)s
"""
SIZE_DISTRIBUTION_TEMPLATE = """
SIZE_DISTRIBUTION    (%(size_distribution)s)
"""


class FluxSimulatorParameters(File):
    _extension = ('.par', )

class FluxSimulatorParametersAndPro(FilePattern):
    _extension = ('.par', '.pro')

def _fluxsimulator_version(execpath):
    cmd = (execpath, '--version')
    m = None
    logfile = tempfile.NamedTemporaryFile()
    try:
        logger.debug(subprocess.list2cmdline(cmd))
        returncode = subprocess.call(cmd,
                                     stderr=logfile,
                                     stdout=logfile )
        # BWA is returning 1
        assert returncode==1, "BWA should have been failing with return code 1 (and it did not)"
        # now dig the information out (note: the subprocess has closed the file - open it again)
        with open(logfile.name) as logfile:
            for row in logfile:
                m = re.match('^Flux-Simulator v([^ ]+)', row)
                if m is not None:
                    break
    except OSError as ose:
        raise unifex.UnifexError("""Command: %s
        %s""" % (' '.join(cmd), ose))
    if m is None:
        raise ValueError('The version number cannot be determined.')
    version = m.groups()[0]
    return version

from railroadtracks import core
class FluxsimulatorExpression(core.StepAbstract):
    _name = 'flux-simulator-expression'
    _default_execpath = 'flux-simulator'
    _version = None
    activities = (ACTIVITY.SIMULATE, )
    class Assets(core.AssetsStep):
        Source = core.assetfactory('Source', [core.AssetAttr('annotation_gtf',
                                                             GFFFile, ''),
                                              core.AssetAttr('genome_dir',
                                                             FilePattern, ''),
                                              core.AssetAttr('parameters',
                                                             FluxSimulatorParameters, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('parameters_and_pro',
                                                             FluxSimulatorParametersAndPro, '')])
    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--NB-MOLECULES', dest='NB_MOLECULES',
                        required=False,
                        type=int)
    parser.add_argument('--force',
                        action = 'store_true')

    def __init__(self, execpath=None):
        if execpath is None:
            self._execpath = self._default_execpath
        else:
            self._execpath = execpath

    @property
    def version(self):
        if self._version is None:
            self._version = _fluxsimulator_version(self._execpath)
        return self._version
        
    def run(self, assets, parameters=()):
        options = self.parser.parse_args(parameters)
        parameters_out = assets.target.parameters_and_pro.name + '.par'
        with open(assets.source.parameters.name) as fh_in, \
             open(parameters_out, 'w') as fh_out:
            logger.debug('Copying parameter file %s to %s.' % (fh_in.name, fh_out.name))
            for row in fh_in:
                if row.startswith('REF_FILE_NAME'):
                    raise ValueError('The parameter file should not define REF_FILE_NAME')
                fh_out.write(row)
            fh_out.write('REF_FILE_NAME\t%s\n' % assets.source.annotation_gtf.name)
            fh_out.write('GEN_DIR\t%s\n' % assets.source.genome_dir.name)
            if options.NB_MOLECULES is not None:
                fh_out.write('NB_MOLECULES\t%i\n' % options.NB_MOLECULES)
        cmd = [self._execpath, '-p', assets.target.parameters_and_pro.name + '.par', '-x', ]
        cmd.extend(parameters)
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd, 
                                               stdout=fnull,
                                               stderr=fnull)
        return (cmd, returncode)


class FluxsimulatorSequencing(core.StepAbstract):
    _name = 'flux-simulator-sequencing'
    _default_execpath = 'flux-simulator'
    _version = None
    activities = (ACTIVITY.SIMULATE, )
    class Assets(core.AssetsStep):
        Source = core.assetfactory('Target', [core.AssetAttr('parameters_and_pro',
                                                             FluxSimulatorParametersAndPro, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('parameters_and_pro',
                                                             FluxSimulatorParametersAndPro, ''),
                                              core.AssetAttr('read1',
                                                             FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('read2',
                                                             FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('lib',
                                                             File, '')])
                                          
    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--read-length',
                        dest = 'read_length',
                        default = 150,
                        type = int,
                        help='Read length')
    parser.add_argument('--read-number',
                        dest = 'read_number',
                        type = int,
                        required = True,
                        help='Total number of reads')
    parser.add_argument('--size-distribution',
                        default = 'N(600,200)',
                        help='Distribution of fragment sizes (default: "%(default)s")')
    parser.add_argument('--error-file',
                        dest = 'error_file',
                        default = '76',
                        type = str,
                        help='Error file (possible values by default are "35" or "76")')
    parser.add_argument('--tmp-dir',
                        dest='tmp_dir',
                        default=None)
    parser.add_argument('--force',
                        action = 'store_true')
    parser.add_argument('--label',
                        required = False)
    
    def __init__(self, execpath=None):
        if execpath is None:
            self._execpath = self._default_execpath
        else:
            self._execpath = execpath
    @property
    def version(self):
        if self._version is None:
            self._version = _fluxsimulator_version(self._execpath)
        return self._version
        
    def run(self, assets, parameters=()):
        options, unknown = self.parser.parse_known_args(parameters)
        parameters_out = assets.target.parameters_and_pro.name + '.par'
        ok_pe = False
        missing_pe = True
        temp_dir = tempfile.mkdtemp()
        syntheticreads_prefix = os.path.join(temp_dir, 'mergedpairs') # is it really a bed file ?
        with open(assets.source.parameters_and_pro.name + '.par') as fh_in, \
             open(parameters_out, 'w') as fh_out:
            logger.debug('Copying parameters from %s to %s.' % (fh_in.name, fh_out.name))
            for row in fh_in:
                if row.startswith('LIB_FILE_NAME'):
                    raise ValueError('The parameter file should not define LIB_FILE_NAME')
                if row.startswith('SEQ_FILE_NAME'):
                    raise ValueError('The parameter file should not define SEQ_FILE_NAME')
                if row.startswith('TMP_DIR') and options.tmp_dir is not None:
                    # silently pass - TMP_DIR is written later
                    pass

                m = re.match('PAIRED_END\t(.+)', row)
                if m is not None:
                    missing_pe = False
                    if m.groups()[0] == 'true':
                        ok_pe = True
                fh_out.write(row)
            fh_out.write('\n')
            fh_out.write('FASTA\tYES\n') # although we want FASTQ, but FluxSimulator wants it that way.
            fh_out.write('READ_NUMBER\t%i\n' % options.read_number)
            fh_out.write('READ_LENGTH\t%i\n' % options.read_length)
            fh_out.write('ERR_FILE\t%s\n' % options.error_file)
            fh_out.write(linesep.join((SEQ_FILE_TEMPLATE,
                                       LIB_FILE_TEMPLATE,
                                       SIZE_DISTRIBUTION_TEMPLATE)) % {'reads_fn': syntheticreads_prefix + '.fastq', 
                                                                          'lib_fn': assets.target.lib.name,
                                                                          'size_distribution': options.size_distribution})
            #fh_out.write(STATS_FILE_TEMPLATE % ({'stats_fn': syntheticreads_prefix + '.txt'}))
            if options.tmp_dir is not None:
                fh_out.write(TMP_DIR_TEMPLATE % ({'tmp_dir': options.tmp_dir}))

            if missing_pe:
                #FIXME: allow paired-end to be False ?
                fh_out.write('PAIRED_END\ttrue\n')
        if not ok_pe and not missing_pe:
            raise ValueError('The input parameter file does not set PAIRED_END to true')
        logger.debug('Copying profile from %s to %s.' % (assets.source.parameters_and_pro.name + '.pro',
                                                         assets.target.parameters_and_pro.name + '.pro'))
        shutil.copy(assets.source.parameters_and_pro.name + '.pro',
                    assets.target.parameters_and_pro.name + '.pro')
        cmd = [self._execpath, '-p', parameters_out, '-l', '-s']
        cmd.extend(parameters)
        #FIXME: channel the ouput into a log file instead ?
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd, stdout=fnull, stderr=fnull)
        _split_mergedpairs((syntheticreads_prefix + '.fastq'), assets.target.read1.name, assets.target.read2.name)
        #os.unlink(tmp_reads.name) # not need - Python will remove it when object collected
        return (cmd, returncode)

#


FluxsimulatorProEntry = collections.namedtuple('FluxsimulatorProEntry',
                                               'locus transcript_id coding length expressed_fraction expressed_number')


class FluxsimulatorPro(object):
    def __init__(self, filename, warn=True):
        if warn and not filename.endswith('.pro'):
            warnings.warn('The file "%s" might not be a PRO file.')
        self.name = filename

    def __iter__(self):
        l = len(FluxsimulatorProEntry._fields)
        with open(self.name) as fh:
            csv_r = csv.reader(fh, delimiter="\t")
            for row in csv_r:
                yield FluxsimulatorProEntry(*(row[:l]))


class FluxsimulatorDerivedExpression(core.StepAbstract):
    """
    Derive new expression values from an existing Flux-Simulator .PRO file by shuffling a proportion
    of expression values.

    The proportion to be shuffled can be modified as a parameter.

    """
    _name = 'flux-simulator-derivedexpression'
    _default_execpath = None
    _version = railroadtracks.__version__
    activities = (ACTIVITY.SIMULATE, )
    COLNAMES = ('Locus', 'Transcript_ID', 'Coding', 'Length', 'Expressed Fraction', 'Expressed Number')

    class Assets(core.AssetsStep):
        Source = core.assetfactory('Source', [core.AssetAttr('parameters_and_pro',
                                                             FluxSimulatorParametersAndPro, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('parameters_and_pro',
                                                             FluxSimulatorParametersAndPro, '')])
                                          
    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--percentage',
                        default = 20,
                        type = float,
                        help='Percentage of expression values to change')
    parser.add_argument('--tmp-dir',
                        dest='tmp_dir',
                        default=None)

    def __init__(self, execpath=None):
        if execpath is None:
            self._execpath = self._default_execpath
        else:
            self._execpath = execpath
    @property
    def version(self):
        return self._version
        
    def run(self, assets, parameters=()):
        options, unknown = self.parser.parse_known_args(parameters)
        parameters_out = assets.target.parameters_and_pro.name + '.par'
        ok_pe = False
        missing_pe = True
        temp_dir = tempfile.mkdtemp()
        syntheticreads_prefix = os.path.join(temp_dir, 'mergedpairs') # is it really a bed file ?
        with open(assets.source.parameters_and_pro.name + '.par') as fh_in, \
             open(parameters_out, 'w') as fh_out:
            logger.debug('Copying parameters from %s to %s.' % (fh_in.name, fh_out.name))
            for row in fh_in:
                if row.startswith('LIB_FILE_NAME'):
                    raise ValueError('The parameter file should not define LIB_FILE_NAME')
                if row.startswith('SEQ_FILE_NAME'):
                    raise ValueError('The parameter file should not define SEQ_FILE_NAME')
                if row.startswith('TMP_DIR') and options.tmp_dir is not None:
                    # silently pass - TMP_DIR is written later
                    pass

                m = re.match('PAIRED_END\t(.+)', row)
                if m is not None:
                    missing_pe = False
                    if m.groups()[0] == 'true':
                        ok_pe = True
                fh_out.write(row)
            if options.tmp_dir is not None:
                fh_out.write(TMP_DIR_TEMPLATE % ({'tmp_dir': options.tmp_dir}))

        if not ok_pe and not missing_pe:
            raise ValueError('The input parameter file does not set PAIRED_END to true')
        logger.debug('Deriving a new profile from %s into %s.' % (assets.source.parameters_and_pro.name + '.pro',
                                                                  assets.target.parameters_and_pro.name + '.pro'))
        # First read it all (easier to swap)
        ef_i = self.COLNAMES.index('Expressed Fraction')
        en_i = self.COLNAMES.index('Expressed Number')
        allrows = list()
        with open(assets.source.parameters_and_pro.name + '.pro') as fh_in:
            csv_r = csv.reader(fh_in, delimiter='\t')
            for n_entries, row in enumerate(csv_r):
                allrows.append((n_entries, row))
        # determine which rows will be modified:
        # 1- get the indices for all rows
        sample = list(range(n_entries))
        # 2- shuffle the indices (we will take the first indices in the shuffle)
        random.shuffle(sample)
        # number of entries modified
        n_modified = int(round(n_entries * options.percentage * 1.0 / 100))
        # require an even number (because we are going to swap the expression values)
        if n_modified % 2 != 0:
            n_modified += 1
        # the first indices will be our sample
        sample = sample[:n_modified]
        
        # iterate through pairs of indices to be swapped
        for i in range(0, n_modified, 2):            
            spl_i = sample[i]
            spl_j = sample[i+1]
            # swap the "Expressed Fraction" (ef_i)
            allrows[spl_i][1][ef_i], allrows[spl_j][1][ef_i] = allrows[spl_j][1][ef_i], allrows[spl_i][1][ef_i]
            # swap the "Expressed Number" (en_i)
            allrows[spl_i][1][en_i], allrows[spl_j][1][en_i] = allrows[spl_j][1][en_i], allrows[spl_i][1][en_i]

        with open(assets.target.parameters_and_pro.name + '.pro', 'w') as fh_out:
            csv_w = csv.writer(fh_out, delimiter='\t')
            csv_w.writerows(row for row_i, row in allrows)
        return (None, 0)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Extension to rnaseqrs to use SGE (no luck with DRMAA, so this is command-line based.
"""
from __future__ import print_function

import os
import re
import csv
import stat
import subprocess
import tempfile
from railroadtracks import hortator, easy, environment, core

import logging
logger = logging.getLogger(__name__)

QSUB_EXEC = 'qsub'

QSUB_JOB_PAT = re.compile('Job ([0-9]+)\\.([0-9]+) exited with exit code ([0-9]+)\\.')

SCRIPT_TEMPLATE = """
#!/bin/bash
#$ -V
%(qsub_preamble)s

# setup
#source %(setup)s

#export PYTHONPATH=%(pythonpath)s:${PYTHONPATH}

# TASKSET_IDS
IFS=', ' read -a TASKSET_IDS_ARRAY <<< ${TASKSET_IDS}
RRT_TASK_ID=${TASKSET_IDS_ARRAY[$((SGE_TASK_ID-1))]}

mkdir -p ${RRT_EXECUTION_DIR}/${RRT_TASK_ID}
cd ${RRT_EXECUTION_DIR}/${RRT_TASK_ID}
python -c 'from railroadtracks.easy import Project; import %(model)s as model; project=Project(model, wd="'${RRT_PROJECT_WD}'"); task=project.get_task('${RRT_TASK_ID}'); task.execute()'
"""


class QsubExecution(object):
    def __init__(self, setup, executiondir, qsub_preamble=''):
        """
        :param setup: BASH code to run before each task is run (setting environment variables common to each task can be done here)
        :type setup: :class:`str`
        :param executiondir: Parent directory where the tasks will run (Hadwrap will create its directories there)
        :param qsub_preamble: Preamble to add to the script called by qsub. This is the place to add qsub configuration statements.
        """
        # Check the presence of "qsub"
        assert environment.Executable.ispresent(QSUB_EXEC)     
        assert os.path.exists(executiondir) and os.path.isdir(executiondir)
        self._executiondir = executiondir
        self.setup = setup
        self.qsub_preamble = qsub_preamble

    def map(self, taskset, tempdir=None):
        assert isinstance(taskset, easy.TaskSet)
        if len(taskset) == 0:
            return
        mapdir = tempfile.mkdtemp(dir=self._executiondir)
        script_path = os.path.join(mapdir, 'script.sh')
        with open(script_path, 'w') as fh_out:
            script = SCRIPT_TEMPLATE % {'qsub_preamble': self.qsub_preamble,
                                        'setup': self.setup,
                                        'pythonpath': os.getenv('PYTHONPATH'),
                                        'model': taskset._project.model.__name__}
            fh_out.write(script)
        
        ts_range = '1-%i' % len(taskset)

        qsub_exec_abs = subprocess.check_output(('which', QSUB_EXEC)).rstrip()
        cmd = (qsub_exec_abs,
               '-t', ts_range,
               '-sync', 'y',
               '-S', '/bin/bash',
               '-o', mapdir,
               '-e', mapdir,
               script_path)
        env = {'RRT_EXECUTION_DIR': mapdir,
               'RRT_PROJECT_WD': taskset._project.wd,
               'TASKSET_IDS': ', '.join(str(x.task_id.id) for x in taskset)}
        for var in ('SGE_CELL',
                    'SGE_EXECD_PORT',
                    'SGE_QMASTER_PORT',
                    'SGE_ROOT',
                    'SGE_CLUSTER_NAME'):
            env[var] = os.getenv(var)

        wd_orig = os.getcwd()
        try:
            os.chdir(mapdir)
            proc = subprocess.Popen(subprocess.list2cmdline(cmd),
                                    env = env,
                                    stdout = subprocess.PIPE,
                                    bufsize = 1,
                                    shell=True)
        finally:
            os.chdir(wd_orig)
        project = taskset._project
        for output in proc.stdout:
            m = QSUB_JOB_PAT.match(output)
            if m is None:
                continue
            jobid, sge_taskid, returncode = m.groups()
            rrt_taskids = tuple(x.task_id for x in taskset)
            if returncode == '0':
                project.get_task(rrt_taskids[int(sge_taskid)]).status = hortator._TASK_DONE
            else:
                project.get_task(rrt_taskids[int(sge_taskid)]).status = hortator._TASK_FAILED
        stdout_value = proc.communicate()[0]



